#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Iterable, Union

import cv2
import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend.argmax import argmax
from fastestimator.backend.concat import concat
from fastestimator.backend.get_image_dims import get_image_dims
from fastestimator.backend.reduce_max import reduce_max
from fastestimator.backend.squeeze import squeeze
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_number

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


@traceable()
class EigenCAM(Trace):
    """A trace which draws EigenCAM heatmaps on top of images.

    These are useful for visualizing the outputs of the feature extractor component of a model. They are relatively
    insensitive to adversarial attacks, so don't use them to try and detect those. See https://arxiv.org/abs/2008.00299
    for more details.

    Args:
        images: The key corresponding to images onto which to draw the CAM outputs.
        activations: The key corresponding to outputs from a convolution layer from which to draw the CAM outputs. You
            can easily extract these from any model by using the 'intermediate_layers' variable in a ModelOp.
        n_components: How many principal components to visualize. If you pass a float between 0 and 1 it will instead
            visualize however many components are required in order to capture the corresponding percentage of the
            variance in the image.
        n_samples: How many images in total to display every epoch (or None to display all available images).
        downsize: Whether to downsize the inputs before svd decomposition in order to speed up processing. If provided,
            the inputs will be proportionally downscaled such that their longest axis length is equal to the `downsize`
            parameter. 64 seems like a good value to try if you are having performance problems.
        labels: The key corresponding to the true labels of the images to be visualized.
        preds: The key corresponding to the model prediction for each image.
        label_mapping: {class_string: model_output_value}.
        outputs: The key into which to write the eigencam images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 images: str,
                 activations: str,
                 n_components: Union[int, float] = 3,
                 n_samples: Optional[int] = 5,
                 downsize: Optional[int] = None,
                 labels: Optional[str] = None,
                 preds: Optional[str] = None,
                 label_mapping: Optional[Dict[str, Any]] = None,
                 outputs: str = "eigencam",
                 mode: Union[None, str, Iterable[str]] = "!train",
                 ds_id: Union[None, str, Iterable[str]] = None):
        self.image_key = images
        self.activation_key = activations
        self.true_label_key = labels
        self.pred_label_key = preds
        inputs = [x for x in (images, activations, labels, preds) if x is not None]
        self.n_components = n_components
        self.n_samples = n_samples
        # TODO - handle non-hashable labels
        self.label_mapping = {val: key for key, val in label_mapping.items()} if label_mapping else None
        self.downsize = downsize
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.images = []
        self.activations = []
        self.labels = []
        self.preds = []
        self.n_found = 0

    def _reset(self) -> None:
        """Clear memory for next epoch.
        """
        self.images = []
        self.activations = []
        self.labels = []
        self.preds = []
        self.n_found = 0

    def _project_2d(self, activations: np.ndarray) -> Tuple[int, List[List[np.ndarray]]]:
        """Project 2D convolution activations maps into 2D principal component maps.

        Args:
            activations: A tensor of shape (batch, channels, height, width) to be transformed.

        Returns:
            (max N_components, Principal component projections of the `activations` (batch x components x image)).
        """
        projections = []
        for activation in activations:
            if self.downsize:
                long_axis = 1 if activation.shape[1] > activation.shape[2] else 2
                long_len = activation.shape[long_axis]
                if long_len > self.downsize:
                    scale = self.downsize / long_len
                    small_activations = []
                    for i in range(activation.shape[0]):
                        small_activations.append(
                            cv2.resize(src=activation[i, ...],
                                       dsize=(int(activation.shape[1]*scale), int(activation.shape[2]*scale)),
                                       interpolation=cv2.INTER_AREA))
                    activation = np.array(small_activations)
            flat = activation.reshape(activation.shape[0], -1).transpose()
            flat = flat - flat.mean(axis=0)
            U, S, VT = np.linalg.svd(flat, full_matrices=True)
            components = []
            n_components = self.n_components
            if isinstance(n_components, float):
                eig_vals = np.square(S)
                pct_explained = np.cumsum(eig_vals) / np.cumsum(eig_vals)[-1]
                n_components = 1 + np.searchsorted(pct_explained, self.n_components)
            for i in range(n_components):
                component_i = flat @ VT[i, :]
                component_i = component_i.reshape(activation.shape[1:])
                components.append(np.maximum(component_i, 0))
            projections.append(components)
        return max([len(x) for x in projections]), projections

    def on_batch_end(self, data: Data) -> None:
        if self.n_samples is None or self.n_found < self.n_samples:
            self.images.append(data[self.image_key])
            self.activations.append(data[self.activation_key])
            if self.true_label_key:
                self.labels.append(data[self.true_label_key])
            if self.pred_label_key:
                self.preds.append(data[self.pred_label_key])
            self.n_found += len(data[self.image_key])

    def on_epoch_end(self, data: Data) -> None:
        # Keep only the user-specified number of samples
        images = concat(self.images)[:self.n_samples or self.n_found]
        _, height, width = get_image_dims(images)
        activations = to_number(concat(self.activations)[:self.n_samples or self.n_found])
        if tf.is_tensor(images):
            activations = np.moveaxis(activations, source=-1, destination=1)  # Activations should be channel first
        args = {}
        labels = None if not self.labels else concat(self.labels)[:self.n_samples or self.n_found]
        if labels is not None:
            if len(labels.shape) > 1:
                labels = argmax(labels, axis=-1)
            if self.label_mapping:
                labels = np.array([self.label_mapping[clazz] for clazz in to_number(squeeze(labels))])
            args[self.true_label_key] = labels
        preds = None if not self.preds else concat(self.preds)[:self.n_samples or self.n_found]
        if preds is not None:
            if len(preds.shape) > 1:
                preds = argmax(preds, axis=-1)
            if self.label_mapping:
                preds = np.array([self.label_mapping[clazz] for clazz in to_number(squeeze(preds))])
            args[self.pred_label_key] = preds
        args[self.image_key] = images
        # Clear memory
        self._reset()
        # Make the image
        n_components, batch_component_image = self._project_2d(activations)
        components = []  # component x image (batch x image)
        for component_idx in range(n_components):
            batch = []
            for base_image, component_image in zip(images, batch_component_image):
                if len(component_image) > component_idx:
                    mask = component_image[component_idx]
                    mask = cv2.resize(mask, (width, height))
                    mask = mask - np.min(mask)
                    mask = mask / np.max(mask)
                    mask = cv2.cvtColor(cv2.applyColorMap(np.uint8(255*mask), cv2.COLORMAP_JET), cv2.COLOR_BGR2RGB)
                    mask = np.float32(mask) / 255
                    # switch to channel first for pytorch
                    if isinstance(base_image, torch.Tensor):
                        mask = np.moveaxis(mask, source=-1, destination=1)
                    new_image = base_image + mask
                    new_image = new_image / reduce_max(new_image)
                else:
                    # There's no component for this image, so display an empty image here
                    new_image = np.ones_like(base_image)
                batch.append(new_image)
            components.append(np.array(batch, dtype=np.float32))

        for idx, elem in enumerate(components):
            args[f"Component {idx}"] = elem

        result = ImgData(**args)
        data.write_without_log(self.outputs[0], result)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar, Union

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def pow(tensor: Tensor, power: Union[int, float, Tensor] ) -> Tensor:
    """Raise a `tensor` to a given `power`.

    This method can be used with Numpy data:
    ```python
    n = np.array([-2, 7, -19])
    b = fe.backend.pow(n, 2)  # [4, 49, 361]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([-2, 7, -19])
    b = fe.backend.pow(t, 2)  # [4, 49, 361]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([-2, 7, -19])
    b = fe.backend.pow(p, 2)  # [4, 49, 361]
    ```

    Args:
        tensor: The input value.
        power: The exponent to raise `tensor` by.

    Returns:
        The exponentiated `tensor`.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        return tf.pow(tensor, power)
    elif isinstance(tensor, torch.Tensor):
        return tensor.pow(power)
    elif isinstance(tensor, np.ndarray):
        return np.power(tensor, power)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.mixed_precision as mixed_precision
import torch

from fastestimator.backend.cast import cast

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def normalize(tensor: Tensor,
              mean: Union[float, Sequence[float]] = (0.485, 0.456, 0.406),
              std: Union[float, Sequence[float]] = (0.229, 0.224, 0.225),
              max_pixel_value: float = 255.0,
              epsilon: float = 1e-7) -> Tensor:
    """
        Compute the normalized value of a `tensor`.

        This method can be used with Numpy data:
        python
        n = np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        b = fe.backend.tensor_normalize(n, 0.5625, 0.2864, 8.0)  # ([[[-1.52752516, -1.0910894 ], [-0.65465364, -0.21821788]], [[ 0.21821788,  0.65465364], [ 1.0910894 ,  1.52752516]]])
        b = fe.backend.tensor_normalize(n, (0.5, 0.625), (0.2795, 0.2795), 8.0) # [[[-1.34164073, -1.34164073], [-0.44721358, -0.44721358]], [[ 0.44721358,  0.44721358], [ 1.34164073,  1.34164073]]]


        This method can be used with TensorFlow tensors:
        python
        t = tf.constant([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        b = fe.backend.tensor_normalize(n, 0.5625, 0.2864, 8.0)  # ([[[-1.52752516, -1.0910894 ], [-0.65465364, -0.21821788]], [[ 0.21821788,  0.65465364], [ 1.0910894 ,  1.52752516]]])
        b = fe.backend.tensor_normalize(n, (0.5, 0.625), (0.2795, 0.2795), 8.0) # [[[-1.34164073, -1.34164073], [-0.44721358, -0.44721358]], [[ 0.44721358,  0.44721358], [ 1.34164073,  1.34164073]]]


        This method can be used with PyTorch tensors:
        python
        p = torch.tensor([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        b = fe.backend.tensor_normalize(n, 0.5625, 0.2864, 8.0)  # ([[[-1.52752516, -1.0910894 ], [-0.65465364, -0.21821788]], [[ 0.21821788,  0.65465364], [ 1.0910894 ,  1.52752516]]])
        b = fe.backend.tensor_normalize(n, (0.5, 0.625), (0.2795, 0.2795), 8.0) # [[[-1.34164073, -1.34164073], [-0.44721358, -0.44721358]], [[ 0.44721358,  0.44721358], [ 1.34164073,  1.34164073]]]

        Args:
            tensor: The input 'tensor' value.
            mean: The mean which needs to applied(eg: 3.8, (0.485, 0.456, 0.406)).
            std: The standard deviation which needs to applied(eg: 3.8, (0.229, 0.224, 0.225)).
            max_pixel_value: The max value of the input data(eg: 255, 65025) to be multipled with mean and std to get actual mean and std.
                            To directly use the mean and std provide set max_pixel_value as 1.
            epsilon: Default value to be added to std to avoid divide by zero error.

        Returns:
            The normalized values of `tensor`.

        Raises:
            ValueError: If `tensor` is an unacceptable data type.
    """
    framework, device = get_framework(tensor)

    mean = get_scaled_data(mean, max_pixel_value, framework, device)
    std = get_scaled_data(std, max_pixel_value, framework, device)

    tensor = (convert_input_precision(tensor) - convert_input_precision(mean)) / (convert_input_precision(std) +
                                                                                  epsilon)

    return tensor


def convert_input_precision(tensor: Tensor) -> Tensor:
    """
        Adjust the input data precision based of environment precision.

        Args:
            tensor: The input value.

        Returns:
            The precision adjusted data(16 bit for mixed precision, 32 bit otherwise).

    """
    precision = 'float32'

    if mixed_precision.global_policy().compute_dtype == 'float16':
        precision = 'float16'

    return cast(tensor, precision)


def get_framework(tensor: Tensor) -> Tuple[str, Optional[str]]:
    """
        Get the framework of the input data.

        Args:
            tensor: The input tensor.
        Returns:
            framework: Framework which is used to load input data.
            device: The device on which the method is executed (Eg. cuda, cpu). Only applicable to torch.
    """
    device = None
    if tf.is_tensor(tensor):
        framework = 'tf'
    elif isinstance(tensor, torch.Tensor):
        framework = 'torch'
        device = tensor.device
    elif isinstance(tensor, np.ndarray):
        framework = 'np'
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))

    return (framework, device)


def get_scaled_data(data: Union[float, Sequence[float]] = (0.485, 0.456, 0.406),
                    scale_factor: float = 255.0,
                    framework: str = 'np',
                    device: Optional[torch.device] = None) -> Tensor:
    """
        Get the scaled value of a input data.

        Args:
            data: The data which needs to be scaled. (eg: 0.4, (0.485, 0.456, 0.406)).
            scale_factor: Scale factor which needs to be multipled with input data.
            framework: Framework currently method is running in.(Eg: 'np','tf', 'torch').
            device: Current device. (eg: 'cpu','cuda').

        Returns:
            The scaled value of input data.
    """
    if framework == 'torch':
        data = torch.tensor(data, device=device)
    elif framework == 'tf':
        data = tf.constant(data)
    else:
        data = np.array(data)

    return data * scale_factor



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from typing import Any, Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from fastestimator.summary import ValWithError
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list, to_number


@traceable()
class CSVLogger(Trace):
    """Log monitored quantities in a CSV file.

    Args:
        filename: Output filename.
        monitor_names: List of keys to monitor. If None then all metrics will be recorded.
        instance_id_key: A key corresponding to data instance ids. If provided, the CSV logger will record per-instance
            metric information into a second csv file.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
    """
    def __init__(self,
                 filename: str,
                 monitor_names: Optional[Union[List[str], str]] = None,
                 instance_id_key: Optional[str] = None,
                 mode: Union[None, str, Iterable[str]] = None) -> None:
        self.instance_id_key = instance_id_key
        monitor_names = to_list(monitor_names)
        instance_id_key = to_list(instance_id_key)
        inputs = monitor_names if monitor_names else ["*"]
        inputs.extend(instance_id_key)
        super().__init__(inputs=inputs, mode=mode)
        self.filename = filename
        self.df_agg = None  # DataFrame for aggregate metrics
        self.df_ins = None  # DataFrame for instance metrics

    def on_begin(self, data: Data) -> None:
        base_keys = ["instance_id", "mode", "step", "epoch"] if self.instance_id_key else ["mode", "step", "epoch"]
        self.df_agg = pd.DataFrame(columns=base_keys)
        self.df_ins = pd.DataFrame(columns=base_keys)

    def on_epoch_end(self, data: Data) -> None:
        tmpdic = {"mode": self.system.mode, "step": self.system.global_step, "epoch": self.system.epoch_idx}

        if "*" in self.inputs:
            for key, value in data.read_logs().items():
                tmpdic[key] = self._parse_val(value)
                if key not in self.df_agg.columns:
                    self.df_agg[key] = ''
        else:
            for key in self.inputs:
                if key == self.instance_id_key:
                    continue
                tmpdic[key] = self._parse_val(data[key])
                if key not in self.df_agg.columns:
                    self.df_agg[key] = ''
        for col in self.df_agg.columns:
            if col not in tmpdic.keys():
                tmpdic[col] = ''

        self.df_agg = self.df_agg.append(tmpdic, ignore_index=True)
        self._save()  # Write on epoch end so that people can see results sooner if debugging

    def on_batch_end(self, data: Data) -> None:
        if self.instance_id_key:
            ins_data = data.read_per_instance_logs()
            if ins_data:
                keys = list((ins_data.keys() if "*" in self.inputs else set(self.inputs)) - {self.instance_id_key})
                ids = data[self.instance_id_key]
                vals = [ins_data.get(key, data.get(key, '')) for key in keys]
                for key in keys:
                    if key not in self.df_ins.columns:
                        self.df_ins[key] = ''
                rows = []
                for sample in zip(ids, *vals):
                    row = {"instance_id": self._parse_val(sample[0]),
                           "mode": self.system.mode,
                           "step": self.system.global_step,
                           "epoch": self.system.epoch_idx,
                           **{key: self._parse_val(val) for key, val in zip(keys, sample[1:])}}
                    for col in self.df_ins.columns:
                        if col not in row.keys():
                            row[col] = ''
                    rows.append(row)
                self.df_ins = self.df_ins.append(rows, ignore_index=True)

        if self.system.mode == "train" and self.system.log_steps and (self.system.global_step % self.system.log_steps
                                                                      == 0 or self.system.global_step == 1):

            tmpdic = {"mode": self.system.mode, "step": self.system.global_step, "epoch": self.system.epoch_idx}

            if "*" in self.inputs:
                for key, value in data.read_logs().items():
                    tmpdic[key] = self._parse_val(value)
                    if key not in self.df_agg.columns:
                        self.df_agg[key] = ''
            else:
                for key in self.inputs:
                    if key == self.instance_id_key:
                        continue
                    tmpdic[key] = self._parse_val(data[key])
                    if key not in self.df_agg.columns:
                        self.df_agg[key] = ''
            for col in self.df_agg.columns:
                if col not in tmpdic.keys():
                    tmpdic[col] = ''

            self.df_agg = self.df_agg.append(tmpdic, ignore_index=True)

    def _save(self) -> None:
        """Write the current state to disk.
        """
        stack = [self.df_ins, self.df_agg]
        if self.system.mode == "test":
            if os.path.exists(self.filename):
                df1 = pd.read_csv(self.filename)
                stack.insert(0, df1)
        stack = pd.concat(stack, axis=0, ignore_index=True)
        stack.to_csv(self.filename, index=False)

    @staticmethod
    def _parse_val(val: Any) -> str:
        """Convert values into string representations.

        Args:
            val: A value to be printed.

        Returns:
            A formatted version of `val` appropriate for a csv file.
        """
        if isinstance(val, str):
            return val
        if isinstance(val, ValWithError):
            return str(val).replace(',', ';')
        val = to_number(val)
        if val.size > 1:
            return np.array2string(val, separator=';')
        if val.dtype.kind in {'U', 'S'}:  # Unicode or String
            # remove the b'' from strings stored in tensors
            return str(val, 'utf-8')
        return str(val)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from typing import Tuple

import numpy as np
import tensorflow as tf
import torch
import torch.nn as nn
import torch.nn.functional as fn
from tensorflow.keras import Sequential, initializers, layers

import fastestimator as fe
from fastestimator.dataset.data import mnist
from fastestimator.op.numpyop.univariate import ExpandDims, Minmax
from fastestimator.op.tensorop.loss import CrossEntropy, L2Regularizaton
from fastestimator.op.tensorop.model import ModelOp, UpdateOp
from fastestimator.trace.metric import Accuracy


class MyNet_torch(torch.nn.Module):
    """A standard DNN implementation in pytorch.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    The MyNet model has 3 dense layers.

    Args:
        input_shape: The shape of the model input (channels, height, width).
        classes: The number of outputs the model should generate.
    """
    def __init__(self, input_shape: Tuple[int, int, int] = (1, 28, 28), classes: int = 10) -> None:
        super().__init__()
        self.fc1 = nn.Linear(input_shape[1] * input_shape[2], 300)
        self.fc1.weight.data.fill_(0.01)
        self.fc1.bias.data.fill_(0.0)

        self.fc2 = nn.Linear(300, 64)
        self.fc2.weight.data.fill_(0.01)
        self.fc2.bias.data.fill_(0.0)

        self.fc3 = nn.Linear(64, classes)
        self.fc3.weight.data.fill_(0.01)
        self.fc3.bias.data.fill_(0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.view(x.size(0), -1)
        x = fn.relu(self.fc1(x))
        x = fn.relu(self.fc2(x))
        x = fn.softmax(self.fc3(x), dim=-1)
        return x


def MyNet_tf(input_shape: Tuple[int, int, int] = (28, 28, 1), classes: int = 10) -> tf.keras.Model:
    """A standard DNN implementation in TensorFlow.

    The MyNet model has 3 dense layers.

    Args:
        input_shape: shape of the input data (height, width, channels).
        classes: The number of outputs the model should generate.

    Returns:
        A TensorFlow MyNet model.
    """
    #_check_input_shape(input_shape)
    model = Sequential()
    model.add(layers.Flatten(input_shape=(28, 28)))
    model.add(
        layers.Dense(units=300,
                     kernel_initializer=initializers.Constant(0.01),
                     bias_initializer=initializers.Zeros(),
                     activation='relu'))
    model.add(
        layers.Dense(units=64,
                     kernel_initializer=initializers.Constant(0.01),
                     bias_initializer=initializers.Zeros(),
                     activation='relu'))
    model.add(
        layers.Dense(units=classes,
                     kernel_initializer=initializers.Constant(0.01),
                     bias_initializer=initializers.Zeros(),
                     activation='softmax'))

    return model


class TestL2Regularization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Weight Decay pytorch model
        cls.beta = 0.1

    def test_l2_regularization_pytorch(self):
        pytorch_l2 = fe.build(model_fn=MyNet_torch, optimizer_fn="adam")
        l2 = L2Regularizaton(inputs='x', outputs='x', model=pytorch_l2)
        output = l2.forward(data=torch.tensor([0.0]), state={})
        output = output.detach()
        self.assertTrue(np.allclose(output.numpy(), 0.12751994))

    def test_l2_regularization_tensorflow(self):
        tf_l2 = fe.build(model_fn=MyNet_tf, optimizer_fn="adam")
        l2 = L2Regularizaton(inputs='x', outputs='x', model=tf_l2)
        output = l2.forward(data=tf.constant([0.0]), state={})
        self.assertTrue(np.allclose(output.numpy(), 0.1275205))

    def test_pytorch_weight_decay_vs_l2(self):
        # Get Data
        train_data, _ = mnist.load_data()
        t_d = train_data.split(128)
        # Initializing models
        pytorch_wd = fe.build(model_fn=MyNet_torch,
                              optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=0.01, weight_decay=self.beta))

        pytorch_l2 = fe.build(model_fn=MyNet_torch, optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=0.01))
        # Initialize pipeline
        pipeline = fe.Pipeline(train_data=t_d,
                               batch_size=128,
                               ops=[ExpandDims(inputs="x", outputs="x", axis=0), Minmax(inputs="x", outputs="x")])
        # Define the two pytorch networks
        network_weight_decay = fe.Network(ops=[
            ModelOp(model=pytorch_wd, inputs="x", outputs="y_pred"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            UpdateOp(model=pytorch_wd, loss_name="ce")
        ])

        network_l2 = fe.Network(ops=[
            ModelOp(model=pytorch_l2, inputs="x", outputs="y_pred"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            L2Regularizaton(inputs="ce", outputs="l2", model=pytorch_l2, beta=self.beta),
            UpdateOp(model=pytorch_l2, loss_name="l2")
        ])

        # defining traces
        traces = [Accuracy(true_key="y", pred_key="y_pred")]

        # Setting up estimators
        estimator_wd = fe.Estimator(pipeline=pipeline,
                                    network=network_weight_decay,
                                    epochs=1,
                                    traces=traces,
                                    train_steps_per_epoch=1)

        estimator_l2 = fe.Estimator(pipeline=pipeline,
                                    network=network_l2,
                                    epochs=1,
                                    traces=traces,
                                    train_steps_per_epoch=1)
        # Training
        print('********************************Pytorch weight decay training************************************')
        estimator_wd.fit()
        print()
        print('********************************Pytorch L2 Regularization training************************************')
        estimator_l2.fit()
        # testing weights
        count = 0
        for wt, l2 in zip(pytorch_wd.parameters(), pytorch_l2.parameters()):
            if ((wt - l2).abs()).sum() < torch.tensor(10**-6):
                count += 1
        self.assertTrue(count == 6)

    def test_pytorch_l2_vs_tensorflow_l2(self):
        # Get Data
        train_data, eval_data = mnist.load_data()
        t_d = train_data.split(128)
        # Initializing Pytorch model
        pytorch_l2 = fe.build(model_fn=MyNet_torch, optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=0.01))
        # Initialize Pytorch pipeline
        pipeline = fe.Pipeline(train_data=t_d,
                        eval_data=eval_data,
                        batch_size=128,
                        ops=[ExpandDims(inputs="x", outputs="x", axis=0),
                                Minmax(inputs="x", outputs="x")])
        # Initialize Pytorch Network
        network_l2 = fe.Network(ops=[
            ModelOp(model=pytorch_l2, inputs="x", outputs="y_pred"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            L2Regularizaton(inputs="ce",outputs="l2",model=pytorch_l2,beta = self.beta),
            UpdateOp(model=pytorch_l2, loss_name="l2")
        ])
        # step 3
        traces = [
            Accuracy(true_key="y", pred_key="y_pred")
        ]
        # Initialize Pytorch estimator
        estimator_l2 = fe.Estimator(pipeline=pipeline,
                            network=network_l2,
                            epochs=1,
                            traces=traces,
                            train_steps_per_epoch=1,
                            monitor_names=["ce","l2"])
        print('********************************Pytorch L2 Regularization training************************************')
        estimator_l2.fit()

        # Converting Pytorch weights to numpy
        torch_wt = []
        for _, param in pytorch_l2.named_parameters():
            if param.requires_grad:
                torch_wt.append(param.detach().numpy())

        # step 1
        pipeline = fe.Pipeline(train_data=t_d,
                               eval_data=eval_data,
                               batch_size=128,
                               ops=[ExpandDims(inputs="x", outputs="x"), Minmax(inputs="x", outputs="x")])
        # step 2
        model_tf = fe.build(model_fn=MyNet_tf, optimizer_fn=lambda: tf.optimizers.SGD(learning_rate=0.01))
        network = fe.Network(ops=[
            ModelOp(model=model_tf, inputs="x", outputs="y_pred"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            L2Regularizaton(inputs="ce",outputs="l2",model=model_tf,beta = self.beta),
            UpdateOp(model=model_tf, loss_name="l2")
        ])
        # step 3
        traces = [
            Accuracy(true_key="y", pred_key="y_pred")
        ]
        estimator = fe.Estimator(pipeline=pipeline,
                                network=network,
                                epochs=1,
                                traces=traces,
                                train_steps_per_epoch=1,
                                monitor_names=["ce","l2"])
        print('*******************************Tensorflow L2 Regularization training***********************************')
        estimator.fit()


        # Converting TF weights to numpy
        tf_wt = []
        for layer in model_tf.layers:
            for w in layer.trainable_variables:
                tf_wt.append(w.numpy())

        # testing weights
        count = 0
        for tf_t,tr in zip(tf_wt,torch_wt):
            if np.sum(np.abs(tf_t-np.transpose(tr))) < (10**-5):
                count += 1
        self.assertTrue(count == 6)



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import random
import unittest

import numpy as np
from nltk.translate.bleu_score import SmoothingFunction, corpus_bleu

import fastestimator as fe
from fastestimator.trace.metric.bleu_score import BleuScore


def inputs():
    while True:
        ground_truth = random.sample(range(1, 100), 10)
        prediction = [i for i in ground_truth if random.random() > 0.3]
        prediction = prediction + [0] * (10 - len(prediction))
        yield {'target_real': ground_truth, 'pred': prediction}


def get_data():
    ground_truth = []
    pred = []
    for _ in range(100):
        data = inputs().__next__()
        ground_truth.append(data['target_real'])
        pred.append(data['pred'])
    return {'target_real': np.array(ground_truth), 'pred': np.array(pred)}


def get_expanded_input(data):
    return np.expand_dims(data, axis=1)


def get_formatted_list(data):
    filtered_data = []
    for i in data:
        filtered_data.append([j for j in i if j != 0])
    return np.array(filtered_data)


class TestBleuScore(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.bleu_score = BleuScore(true_key="target_real",
                                    pred_key="pred",
                                    output_name="bleu_score",
                                    n_gram=4,
                                    mode="!train",
                                    per_ds=False)
        self.train_data = get_data()
        print(self.train_data['target_real'].shape)
        print(self.train_data['pred'].shape)

    def get_batch_data(self, batch_size):
        batch_data = []
        for i in range(0, 100, batch_size):
            gt = self.train_data['target_real'][i:i + batch_size]
            pred = self.train_data['pred'][i:i + batch_size]
            batch_data.append(fe.util.Data({'target_real': np.array(gt), 'pred': np.array(pred)}))
        return batch_data

    def test_weights(self):
        weights = self.bleu_score.weights
        self.assertEqual(weights, (0.25, 0.25, 0.25, 0.25))

    def test_invalid_ngram(self):
        with self.assertRaises(ValueError):
            _ = BleuScore(true_key="target_real", pred_key="pred", output_name="bleu_score", n_gram=-1, mode="!train")

    def test_corpus_bleu_score(self):

        batch_size = 4
        for batch_data in self.get_batch_data(batch_size):
            self.bleu_score.on_batch_end(batch_data)

        predicited_bleu_score = self.bleu_score.get_corpus_bleu_score()
        expected_bleu_score = corpus_bleu(get_expanded_input(self.train_data['target_real']),
                                          get_formatted_list(self.train_data['pred']),
                                          self.bleu_score.weights,
                                          smoothing_function=SmoothingFunction().method4)
        self.assertAlmostEqual(predicited_bleu_score, expected_bleu_score, delta=0.001)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
import gc
import multiprocessing as mp
import os
import time
from copy import deepcopy
from operator import mul
from threading import Lock
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union

import numpy as np
import tensorflow as tf
from torch.utils.data import DataLoader, Dataset

from fastestimator.backend.to_tensor import to_tensor
from fastestimator.dataset.dataloader import FEDataLoader
from fastestimator.dataset.op_dataset import OpDataset, _DelayedDeepDict
from fastestimator.op.numpyop.meta.fuse import Fuse
from fastestimator.op.numpyop.meta.one_of import OneOf
from fastestimator.op.numpyop.meta.repeat import Repeat
from fastestimator.op.numpyop.meta.sometimes import Sometimes
from fastestimator.op.numpyop.numpyop import NumpyOp, forward_numpyop, Batch
from fastestimator.schedule.schedule import Scheduler, get_current_items, EpochScheduler, \
    RepeatScheduler
from fastestimator.util.data import FilteredData
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import get_num_devices, to_list, to_set

DataSource = TypeVar('DataSource', Dataset, DataLoader, tf.data.Dataset)


@traceable(blacklist=('ctx_loader', 'ctx_lock'))
class Pipeline:
    """A data pipeline class that takes care of data pre-processing.

    Args:
        train_data: The training data, or None if no training data is available.
        eval_data: The evaluation data, or None if no evaluation data is available.
        test_data: The testing data, or None if no evaluation data is available.
        batch_size: The batch size to be used by the pipeline. If the batch_size is also set by a Batch Op, that value
            will take precedence over this one (for example, if you want to set the batch_size based on mode or ds_is).
            NOTE: This argument is only applicable when using a FastEstimator Dataset.
        ops: NumpyOps to be used for pre-processing. NOTE: This argument is only applicable when using a FastEstimator
            Dataset.
        num_process: Number of CPU threads to use for data pre-processing. NOTE: This argument is only applicable when
            using a FastEstimator Dataset. None will default to min(n_cpus, max(32, 32*n_gpus)). Multiprocessing can be
            disabled by passing 0 here, which can be useful for debugging.
    """
    ops: List[Union[NumpyOp, Scheduler[NumpyOp]]]
    data: Dict[str, Dict[Optional[str], Union[DataSource, Scheduler[DataSource]]]]  # {"mode": {"ds_id": ds}}

    def __init__(self,
                 train_data: Union[None,
                                   DataSource,
                                   Scheduler[DataSource],
                                   Dict[str, Union[DataSource, Scheduler[DataSource]]]] = None,
                 eval_data: Union[None, DataSource, Scheduler[DataSource], Dict[str, DataSource]] = None,
                 test_data: Union[None, DataSource, Scheduler[DataSource], Dict[str, DataSource]] = None,
                 batch_size: Union[None, int, Scheduler[int]] = None,
                 ops: Union[None, NumpyOp, Scheduler[NumpyOp], List[Union[NumpyOp, Scheduler[NumpyOp]]]] = None,
                 num_process: Optional[int] = None):
        data = {x: y for (x, y) in zip(["train", "eval", "test"], [train_data, eval_data, test_data]) if y}
        self.data = self._register_ds_ids(data)
        self.batch_size = batch_size
        self.ops = to_list(ops)
        if mp.get_start_method(allow_none=True) is None and os.name != 'nt':
            mp.set_start_method('fork')
        if mp.get_start_method(allow_none=True) != 'fork':
            print("FastEstimator-Warn: Pipeline multiprocessing is disabled. OS must support the 'fork' start method.")
            num_process = 0
        self.num_process = num_process if num_process is not None else min(os.cpu_count(), 32 * get_num_devices())
        self._verify_inputs(**{k: v for k, v in locals().items() if k != 'self'})
        # Loader Variables
        self.ctx_lock = Lock()
        self.ctx_mode = 'train'
        self.ctx_epoch = 1
        self.ctx_shuffle = True
        self.ctx_output_keys = None
        self.ctx_loader = None
        self.ctx_ds_id = None
        self.ctx_batch_size = None
        self.ctx_ops = []
        self.ctx_batch_info = Batch()
        self.ctx_batch_ops = []
        self.ctx_batch_input_keys = set()

    @staticmethod
    def _register_ds_ids(
            data: Dict[
                str, Union[DataSource, Scheduler[DataSource], Dict[str, Union[DataSource, Scheduler[DataSource]]]]]
    ) -> Dict[str, Dict[Optional[str], Union[DataSource, Scheduler[DataSource]]]]:
        """Associate dataset of each mode with a `ds_id`.

        Args:
            data: A dictionary with mode as key, dataset as value.
        """
        forbidden_ds_id_chars = {":", "!", ";", "|"}
        for mode, dataset in data.items():
            if isinstance(dataset, dict):
                for ds_name in dataset:
                    assert isinstance(ds_name, str) and len(ds_name) > 0, \
                        "dataset id must be a string, found {}".format(ds_name)
                    assert not any(char in ds_name for char in forbidden_ds_id_chars), \
                        "dataset id should not contain forbidden characters like ':', ';', '!', '|', " + \
                        "found {} in pipeline".format(ds_name)
            else:
                # Empty string is special, matches against ops which require '!ds1' but not 'ds1'
                data[mode] = {"": dataset}
        return data

    def _verify_inputs(self, **kwargs) -> None:
        """A helper method to ensure that the Pipeline inputs are valid.

        Args:
            **kwargs: A collection of variable / value pairs to validate.

        Raises:
            AssertionError: If `batch_size`, `ops`, or `num_process` were specified in the absence of a FastEstimator
                Dataset.
        """
        fe_dataset = False
        for dataset in get_current_items(set(d for ds in self.data.values() for d in ds.values())):
            fe_dataset = self._verify_dataset(dataset, **kwargs) or fe_dataset
        if not fe_dataset:
            assert kwargs['batch_size'] is None, "Pipeline only supports batch_size with built-in (FE) datasets"
            assert kwargs['ops'] is None, "Pipeline only supports ops with built-in (FE) datasets"
            assert kwargs['num_process'] is None, "Pipeline only support num_process with built-in (FE) datasets"
        # Make sure that the user provides at most 1 Batch Op for a given epoch/mode/ds_id
        batch_ops = []
        schedule_epochs = {1}
        schedule_cycles = set()
        for op in self.ops:
            if isinstance(op, Batch):
                batch_ops.append(op)
            if isinstance(op, Scheduler):
                # Only keep the scheduler if it contains at least one Batch op
                vals = op.get_all_values()
                for val in vals:
                    if isinstance(val, Batch):
                        batch_ops.append(op)
                        if isinstance(op, EpochScheduler):
                            schedule_epochs |= op.epoch_dict.keys()
                        elif isinstance(op, RepeatScheduler):
                            schedule_cycles.add(op.cycle_length)
                        else:
                            # Some unknown scheduler, no known shortcuts so just try first 100 epochs to be safe
                            schedule_epochs |= {*range(1, 100)}
                        break
        # After m*n steps all possible m and n combinations will be visited
        schedule_cycles = functools.reduce(mul, schedule_cycles, 1)
        # Consider x + m*n epochs for each epoch scheduler x value
        schedule_epochs = sorted({epoch for base_epoch in schedule_epochs for epoch in
                                  list(range(base_epoch, base_epoch + schedule_cycles))})
        for mode, id_ds in self.data.items():
            for ds_id in id_ds.keys():
                for epoch in schedule_epochs:
                    ops = get_current_items(batch_ops, run_modes=mode, epoch=epoch, ds_id=ds_id)
                    # We have to do an instance check again since the user could technically use a scheduler that has a
                    # Batch Op at one point, but some other Op (or None) at a different point
                    ops = [op for op in ops if isinstance(op, Batch)]
                    assert len(ops) < 2, "You may provide at most 1 batch op for a given epoch/mode/ds_id combination"

    def _verify_dataset(self, dataset: DataSource, **kwargs) -> bool:
        """A helper function to ensure that all of a dataset's arguments are correct.

        Args:
            dataset: The dataset to validate against.
            **kwargs: A selection of variables and their values which must be validated.

        Returns:
            True iff the `dataset` is a PyTorch Dataset (as opposed to a DataLoader or tf.data.Dataset).

        Raises:
            AssertionError: If the `kwargs` are found to be invalid based on the given `dataset`.
            ValueError: If the `dataset` is of an unknown type.
        """
        if isinstance(dataset, Dataset):
            # batch_size check
            for batch_size in get_current_items(to_list(self.batch_size)):
                assert isinstance(batch_size, int), "unsupported batch_size format: {}".format(type(batch_size))
            # ops check
            for op in get_current_items(self.ops):
                assert isinstance(op, NumpyOp), "unsupported op format, must provide NumpyOp in Pipeline"
            # num_process check
            assert isinstance(self.num_process, int), "number of processes must be an integer"
            return True
        elif isinstance(dataset, (DataLoader, tf.data.Dataset)):
            if kwargs['batch_size'] is not None:
                print("FastEstimator-Warn: batch_size will only be used for built-in dataset")
            if kwargs['ops'] is not None:
                print("FastEstimator-Warn: ops will only be used for built-in dataset")
            if kwargs['num_process'] is not None:
                print("FastEstimator-Warn: num_process will only be used for built-in dataset")
            return False
        else:
            raise ValueError("Unsupported dataset type: {}".format(type(dataset)))

    def _get_op_split(self, mode: str, epoch: int, ds_id: str) -> Tuple[List[NumpyOp], Batch, List[NumpyOp]]:
        """Figure out which ops are pre-batch vs post-batch.

        Args:
            mode: The current mode.
            epoch: The current epoch.
            ds_id: The current dataset.

        Returns:
            (instance ops, batch info, batch ops).
        """
        batch_info = Batch()
        instance_ops = []
        batch_ops = []
        ops = get_current_items(self.ops, run_modes=mode, epoch=epoch, ds_id=ds_id)
        target = instance_ops
        for op in ops:
            if isinstance(op, Batch):
                batch_info = op
                target = batch_ops
                continue
            target.append(op)
        return instance_ops, batch_info, batch_ops

    def get_modes(self, epoch: Optional[int] = None) -> Set[str]:
        """Get the modes for which the Pipeline has data.

        Args:
            epoch: The current epoch index

        Returns:
            The modes for which the Pipeline has data.
        """
        if epoch is None:
            all_modes = set(self.data.keys())
        else:
            all_modes = []
            for mode, datasets in self.data.items():
                for dataset in datasets.values():
                    if isinstance(dataset, Scheduler):
                        dataset = dataset.get_current_value(epoch)
                    if dataset:
                        all_modes.append(mode)
        return to_set(all_modes)

    def get_ds_ids(self, epoch: int, mode: str) -> List[Union[str, None]]:
        """Get the ds_ids for a given epoch and mode.

        Args:
            epoch: The current epoch index.
            mode: The current execution mode.

        Returns:
            The ds_ids of the current epoch and mode.
        """
        ds_ids = []
        if mode in self.data:
            datasets = self.data[mode]
            for ds_id, dataset in datasets.items():
                if isinstance(dataset, Scheduler):
                    dataset = dataset.get_current_value(epoch)
                if dataset:
                    ds_ids.append(ds_id)
        return ds_ids

    def benchmark(self,
                  mode: str = "train",
                  epoch: int = 1,
                  ds_id: Optional[str] = None,
                  num_steps: int = 1000,
                  log_interval: int = 100,
                  detailed: bool = True) -> None:
        """Benchmark the pipeline processing speed.

        Args:
            mode: The execution mode to benchmark. This can be 'train', 'eval' or 'test'.
            epoch: The epoch index to benchmark. Note that epoch indices are 1-indexed.
            ds_id: The ds_id to benchmark. If None, all ds_ids will be benchmarked.
            num_steps: The number of steps over which to perform the benchmark.
            log_interval: The logging interval.
            detailed: Whether to display the detailed time used by each operator.
        """
        if ds_id is None:
            ds_ids = self.get_ds_ids(epoch=epoch, mode=mode)
        else:
            ds_ids = [ds_id]

        for ds_id in ds_ids:
            with self(mode=mode, epoch=epoch, ds_id=ds_id, steps_per_epoch=num_steps) as loader:
                if isinstance(loader, tf.data.Dataset):
                    loader = loader.take(num_steps)
                start = time.perf_counter()
                for idx, _ in enumerate(loader, start=1):
                    if idx % log_interval == 0:
                        duration = time.perf_counter() - start
                        iters_per_sec = log_interval / duration
                        ds_str = f"Dataset: {ds_id}, " if ds_id else ""
                        print("FastEstimator-Benchmark ({}): {}Step: {}, Epoch: {}, Steps/sec: {}".format(
                            mode.capitalize(), ds_str, idx, epoch, iters_per_sec))
                        start = time.perf_counter()
                # Pipeline Operations Benchmarking when using FEDataset
                if isinstance(loader, FEDataLoader) and isinstance(loader.dataset, OpDataset) and detailed:
                    # (n_visited, duration)
                    duration_list = np.zeros(shape=(len(self.ctx_ops) + 1 + len(self.ctx_batch_ops), 2))
                    data_len = len(loader.dataset)
                    ds_str = f", Dataset: {ds_id}" if ds_id else ""
                    print("\nBreakdown of time taken by Pipeline Operations (Mode: {}, Epoch: {}{})\n".format(
                        mode.capitalize(), epoch, ds_str))
                    extra_memory_management_time = 0
                    for _ in range(log_interval):
                        filtered = False
                        batch = []
                        index = np.random.randint(data_len)
                        items = deepcopy(loader.dataset.dataset[index])
                        if isinstance(items, list):
                            while not batch:
                                filtered = False
                                # BatchDataset may randomly sample the same elements multiple times, avoid reprocessing
                                unique_samples = set()
                                for item in items:
                                    if id(item) not in unique_samples:
                                        for i, op in enumerate(self.ctx_ops):
                                            start = time.perf_counter()
                                            op_data = forward_numpyop([op], item, {'mode': loader.dataset.mode})
                                            duration = time.perf_counter() - start
                                            duration_list[i][0] += 1
                                            duration_list[i][1] += duration
                                            if isinstance(op_data, FilteredData):
                                                filtered = True
                                                break
                                        unique_samples.add(id(item))
                                if not filtered:
                                    batch = items
                        else:
                            while len(batch) < (self.ctx_batch_size or 1):
                                filtered = False
                                for i, op in enumerate(self.ctx_ops):
                                    start = time.perf_counter()
                                    op_data = forward_numpyop([op], items, {'mode': mode})
                                    duration = time.perf_counter() - start
                                    duration_list[i][0] += 1
                                    duration_list[i][1] += duration
                                    if isinstance(op_data, FilteredData):
                                        filtered = True
                                        break
                                if not filtered:
                                    batch.append(items)
                                index = np.random.randint(data_len)
                                items = deepcopy(loader.dataset.dataset[index])
                        if not filtered:
                            # Perform the batching
                            start = time.perf_counter()
                            batch = self.ctx_batch_info.collate_fn(batch)
                            duration = time.perf_counter() - start
                            duration_list[len(self.ctx_ops)][0] += 1
                            duration_list[len(self.ctx_ops)][1] += duration
                            # Perform batch ops
                            start = time.perf_counter()
                            # Transform to numpy to not bias against the first op in the batch_op chain
                            batch = to_tensor(batch, target_type='np')
                            extra_memory_management_time += time.perf_counter() - start

                            for i, op in enumerate(self.ctx_batch_ops, start=len(self.ctx_ops) + 1):
                                start = time.perf_counter()
                                op_data = forward_numpyop([op], data=batch, state={'mode': mode}, batched='np')
                                duration = time.perf_counter() - start
                                duration_list[i][0] += 1
                                duration_list[i][1] += duration
                                if isinstance(op_data, FilteredData):
                                    break
                            # Count extra time needed to cast data back to torch
                            start = time.perf_counter()
                            to_tensor(batch, target_type='torch', shared_memory=True)
                            extra_memory_management_time += time.perf_counter() - start

                    if self.ctx_batch_ops:
                        # Extra memory management penalty is only incurred when using batch ops
                        duration_list[len(self.ctx_ops)][1] += extra_memory_management_time

                    total_time = np.sum(duration_list[:, 1])
                    normalized_times_ms = 1000 * duration_list[:, 1] / np.maximum(duration_list[:, 0], 1)
                    op_names = ["Op"]

                    for op in self.ctx_ops + [self.ctx_batch_info] + self.ctx_batch_ops:
                        if isinstance(op, Sometimes) and op.op:
                            op_names.append(op.__class__.__name__ + " (" + op.op.__class__.__name__ + ")")
                        elif isinstance(op, Repeat) and op.op:
                            op_names.append(op.__class__.__name__ + " (" + op.op.__class__.__name__ + ")")
                        elif isinstance(op, OneOf) and op.ops:
                            op_names.append(op.__class__.__name__ + " (" +
                                            ", ".join([sub_op.__class__.__name__ for sub_op in op.ops]) + ")")
                        elif isinstance(op, Fuse) and op.ops:
                            op_names.append(op.__class__.__name__ + " (" +
                                            ", ".join([sub_op.__class__.__name__ for sub_op in op.ops]) + ")")
                        elif isinstance(op, Batch):
                            op_names.append("<Collating Batch>")
                        else:
                            op_names.append(op.__class__.__name__)

                    max_op_len = max(len(op_name) for op_name in op_names)
                    max_in_len = max([len(", ".join(op.inputs)) for op in
                                      self.ctx_ops + [self.ctx_batch_info] + self.ctx_batch_ops] + [len("Inputs")])
                    max_out_len = max([len(", ".join(op.outputs)) for op in
                                       self.ctx_ops + [self.ctx_batch_info] + self.ctx_batch_ops] + [len("Outputs")])
                    ms_visit_len = max(len("{:.3f}".format(max(normalized_times_ms))), len("ms / Visit"))
                    visit_len = max(len(f"{int(np.max(duration_list[:, 0]))}"), len("Visits"))

                    print("{}: {}: {}: {}: {}: {}".format("Op".ljust(max_op_len + 1),
                                                          "Inputs".ljust(max_in_len + 1),
                                                          "Outputs".ljust(max_out_len + 1),
                                                          "ms / Visit".ljust(ms_visit_len + 1),
                                                          "Visits".ljust(visit_len + 1),
                                                          "Time (Total)".rjust(12)))
                    print("-" * (max_op_len + max_in_len + max_out_len + visit_len + 37))
                    for i, op in enumerate(self.ctx_ops + [self.ctx_batch_info] + self.ctx_batch_ops):
                        print("{}: {}: {}: {}: {}: {:11.2f}%".format(
                            op_names[i + 1].ljust(max_op_len + 1),
                            ", ".join(op.inputs).ljust(max_in_len + 1),
                            ", ".join(op.outputs).ljust(max_out_len + 1),
                            "{:.3f}".format(normalized_times_ms[i]).ljust(ms_visit_len + 1),
                            str(int(duration_list[i][0])).ljust(visit_len + 1),
                            100 * duration_list[i][1] / total_time))
                    if self.ctx_batch_ops:
                        penalty = round(100*(duration_list[len(self.ctx_ops)][1] - extra_memory_management_time) /
                                        duration_list[len(self.ctx_ops)][1], 1)
                        print(f"\nNote that collation time would be cut by ~{penalty}% if there were no batched ops.")
                print("\n")  # to make printing more obvious

    def get_scheduled_items(self, mode: str) -> List[Any]:
        """Get a list of items considered for scheduling.

        Args:
            mode: Current execution mode.

        Returns:
            List of schedulable items in Pipeline.
        """
        all_items = self.ops + [self.batch_size] + list(self.data[mode].values())
        return all_items

    def get_epochs_with_data(self, total_epochs: int, mode: str) -> Set[int]:
        """Get a set of epoch indices that contains data given mode.

        Args:
            total_epochs: Total number of epochs.
            mode: Current execution mode.

        Returns:
            Set of epoch indices.
        """
        epochs_with_data = set()
        datasets = self.data[mode]
        for dataset in datasets.values():
            if isinstance(dataset, Scheduler):
                epochs_with_data_ds = set(epoch for epoch in range(1, total_epochs + 1)
                                          if dataset.get_current_value(epoch))
                epochs_with_data = epochs_with_data | epochs_with_data_ds
            elif dataset:
                epochs_with_data_ds = set(range(1, total_epochs + 1))
                epochs_with_data = epochs_with_data | epochs_with_data_ds
                break
        return epochs_with_data

    def transform(self,
                  data: Dict[str, Any],
                  mode: str,
                  epoch: int = 1,
                  ds_id: str = '',
                  target_type: str = 'np') -> Union[Dict[str, Any], FilteredData]:
        """Apply all pipeline operations on a given data instance for the specified `mode` and `epoch`.

        Args:
            data: Input data in dictionary format.
            mode: The execution mode in which to run. This can be "train", "eval", "test" or "infer".
            epoch: The epoch index to run. Note that epoch indices are 1-indexed.
            ds_id: The current dataset id.
            target_type: What kind of tensor(s) to create. One of "tf", "torch", or "np".

        Returns:
            The transformed data.
        """
        data = deepcopy(data)
        instance_ops, batch_spec, batch_ops = self._get_op_split(mode=mode, epoch=epoch, ds_id=ds_id)
        state = {'mode': mode}
        op_data = forward_numpyop(instance_ops, data, state)
        if isinstance(op_data, FilteredData):
            return op_data
        data = batch_spec.collate_fn([data])
        op_data = forward_numpyop(batch_ops, data, state, batched='torch')
        if isinstance(op_data, FilteredData):
            return op_data
        return to_tensor(data, target_type=target_type)

    def get_results(self,
                    mode: str = "train",
                    epoch: int = 1,
                    ds_id: str = '',
                    num_steps: int = 1,
                    shuffle: bool = False) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Get sample Pipeline outputs.

        Args:
            mode: The execution mode in which to run. This can be "train", "eval", or "test".
            epoch: The epoch index to run. Note that epoch indices are 1-indexed.
            num_steps: Number of steps (batches) to get.
            shuffle: Whether to use shuffling.
            ds_id: The current dataset id.

        Returns:
            A list of batches of Pipeline outputs.
        """
        results = []
        with self(mode=mode, epoch=epoch, ds_id=ds_id, shuffle=shuffle) as loader:
            if isinstance(loader, tf.data.Dataset):
                loader = loader.take(num_steps)
            if loader:
                for idx, batch in enumerate(loader, start=1):
                    results.append(batch)
                    if idx == num_steps:
                        break
                if len(results) == 1:
                    results = results[0]
            return results

    def __call__(self,
                 mode: str,
                 epoch: int = 1,
                 ds_id: str = '',
                 shuffle: Optional[bool] = None,
                 steps_per_epoch: Optional[int] = None,
                 output_keys: Optional[Set[str]] = None) -> 'Pipeline':
        """Prepare this Pipeline for a given `mode` and `epoch`.

        A given pipeline can only provide one loader at a time. This helps to prevent issues with multi-threading.

        ```python
        pipe = Pipeline(...)
        with pipe(mode='eval', epoch=2) as loader:
            for batch in loader:
                print(batch)
        ```

        Args:
            mode: The execution mode for the loader. This can be 'train', 'eval' or 'test'.
            epoch: The epoch index for the loader. Note that epoch indices are 1-indexed.
            ds_id: The dataset id to consider for the loader.
            shuffle: Whether to shuffle the data. If None, the value for shuffle is based on mode. NOTE: This argument
                is only used with FastEstimator Datasets.
            steps_per_epoch: Training or Evaluation will be cut short or extended to complete N steps even if loader is
                not yet exhausted. If None, all data will be used.
            output_keys: What keys can be produced from pipeline. If None or empty, all keys will be considered.

        Returns:
            The pipeline, but with `mode` and `epoch` set for use in a loader.

        Raises:
            ValueError: If called while the pipeline already has an active loader.
        """
        # Make sure that a loader isn't currently instantiated with other settings
        acquired = self.ctx_lock.acquire(blocking=False)
        if not acquired:
            raise ValueError("You cannot invoke a Pipeline's __call__ method while it already has an active loader.")
        self.ctx_mode = mode
        self.ctx_epoch = epoch
        self.ctx_ds_id = ds_id
        self.ctx_shuffle = mode == 'train' if shuffle is None else shuffle
        self.ctx_steps_per_epoch = steps_per_epoch
        self.ctx_output_keys = output_keys or set()
        self.ctx_ops, self.ctx_batch_info, self.ctx_batch_ops = self._get_op_split(mode=mode, epoch=epoch, ds_id=ds_id)
        # Figure out which input keys are required by the batch ops (so they don't get pruned too early)
        self.ctx_batch_input_keys = set()
        batch_produced_keys = set()
        for op in get_current_items(self.ctx_batch_ops, mode, epoch, ds_id=ds_id):
            self.ctx_batch_input_keys.update(set(key for key in op.inputs if key not in batch_produced_keys))
            batch_produced_keys.update(op.outputs)
        # Decide on the batch size (this might still be ignored later if the user is using a BatchDataset)
        self.ctx_batch_size = self.ctx_batch_info.batch_size
        if self.ctx_batch_size is None:
            # batch size
            batch_size = self.batch_size
            if isinstance(batch_size, Scheduler):
                batch_size = batch_size.get_current_value(self.ctx_epoch)
            self.ctx_batch_size = batch_size
        self.ctx_lock.release()
        return self

    def __enter__(self) -> Union[DataLoader, tf.data.Dataset]:
        """Get a data loader from the Pipeline for the current epoch and mode.

        A given pipeline can only provide one loader at a time. This helps to prevent issues with multi-threading.

        ```python
        pipe = Pipeline(...)
        with pipe(mode='eval', epoch=2) as loader:
            for batch in loader:
                print(batch)
        ```

        Returns:
            A data loader for the current `mode` and `epoch`.

        Raises:
            ValueError: If called while the pipeline already has an active loader.
        """
        acquired = self.ctx_lock.acquire(blocking=False)
        if not acquired:
            raise ValueError("You cannot generate a new loader from this Pipeline before closing its other loader.")
        # Release the lock if arguments are invalid so that people in Jupyter / debug consoles don't get stuck
        if self.ctx_mode not in self.data:
            self.ctx_lock.release()
            raise KeyError(f"Pipeline has no data for mode '{self.ctx_mode}'")
        if self.ctx_ds_id not in self.data[self.ctx_mode]:
            self.ctx_lock.release()
            raise KeyError(f"The dataset id '{self.ctx_ds_id}' is not present in {self.ctx_mode} mode")
        data = self.data[self.ctx_mode][self.ctx_ds_id]
        if isinstance(data, Scheduler):
            data = data.get_current_value(self.ctx_epoch)
        if isinstance(data, Dataset):
            # Results will be immediately converted to tensors, so don't need deep_remainder
            op_dataset = OpDataset(data,
                                   self.ctx_ops,
                                   self.ctx_mode,
                                   self.ctx_output_keys | self.ctx_batch_input_keys if self.ctx_output_keys else None,
                                   deep_remainder=False)
            # check whether to batch the data
            batch_size = None if op_dataset.fe_batch else self.ctx_batch_size
            # Figure out whether a postprocessing function is needed (for batched ops)
            postprocess_fn = None
            if self.ctx_batch_ops:
                postprocess_fn = functools.partial(_batch_postprocess,
                                                   ops=self.ctx_batch_ops,
                                                   output_keys=self.ctx_output_keys,
                                                   mode=self.ctx_mode)
            try:
                data = FEDataLoader(op_dataset,
                                    postprocess_fn=postprocess_fn,
                                    batch_size=batch_size,
                                    shuffle=self.ctx_shuffle,
                                    steps_per_epoch=self.ctx_steps_per_epoch,
                                    num_workers=self.num_process,
                                    drop_last=self.ctx_batch_info.drop_last,
                                    collate_fn=self.ctx_batch_info.collate_fn)
            except ValueError as err:
                self.ctx_lock.release()
                raise err
            self.ctx_loader = data
        return data

    def __exit__(self, *exc: Tuple[Optional[Type], Optional[Exception], Optional[Any]]) -> None:
        if self.ctx_loader is not None:
            self.ctx_loader.shutdown()
            self.ctx_loader = None
        # Manually triggering gc here seems to be necessary in order to avoid problems with repeated invocations of FE
        # killing one another through multi-processing.
        gc.collect()
        self.ctx_lock.release()


def _batch_postprocess(data: Dict[str, Any], ops: List[NumpyOp], output_keys: Set[str], mode: str) -> \
        Union[Dict[str, Any], FilteredData]:
    op_data = forward_numpyop(ops=ops, data=data, state={'mode': mode}, batched='torch')
    if isinstance(op_data, FilteredData):
        return op_data
    if output_keys:
        for key in data.keys() - output_keys:
            if key not in _DelayedDeepDict.warned:
                _DelayedDeepDict.warned.add(key)
                print("FastEstimator-Warn: the key '{}' is being pruned since it is unused outside of the Pipeline."
                      " To prevent this, you can declare the key as an input of a Trace or TensorOp.".format(key))
            data.pop(key)
    return data



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
import random
from abc import ABC
from typing import Any, Callable, Dict, List, Optional, Sized, Tuple, Union

import numpy as np
from torch import Tensor
from torch.utils.data import DataLoader, Dataset, Sampler, _DatasetKind
from torch.utils.data._utils.collate import default_collate, default_convert
from torch.utils.data._utils.fetch import _MapDatasetFetcher
from torch.utils.data.dataloader import _BaseDataLoaderIter, _MultiProcessingDataLoaderIter
from torch.utils.data.dataloader import _SingleProcessDataLoaderIter

from fastestimator.dataset.extend_dataset import ExtendDataset
from fastestimator.dataset.op_dataset import OpDataset
from fastestimator.util.data import FilteredData


class FEDataLoader(DataLoader):
    """A Data Loader that can handle filtering data.

    This class is intentionally not @traceable.

    Args:
        dataset: The dataset to be drawn from. The dataset may optionally implement .fe_reset_ds(bool) and/or
            .fe_batch_indices(int) methods to modify the system's sampling behavior. See fe.dataset.BatchDataset for an
            example which uses both of these methods.
        postprocess_fn: A function to run on a collated batch of data before returning it. This function can return a
            FilteredData object in order to drop the given batch.
        batch_size: The batch size to use (or None if the dataset is already providing a batch).
        steps_per_epoch: How many steps to have per epoch. If None the loader will perform a single pass through the
            dataset (unless samples are filtered with replacement, in which case the dataset may be passed over multiple
            times). If `steps_per_epoch` is set, it will truncate or expand the dataset until the specified number of
            steps are reached. When expanding datasets, they will be exhausted in their entirety before being
            re-sampled, equivalent to running multiple epochs of training one after the other (unless you are also
            filtering data, in which case at most one batch of data might be seen after the re-shuffling occurs).
        shuffle: Whether to shuffle the dataset.
        num_workers: How many multiprocessing threads to use (unix/mac only).
        collate_fn: What function to use to collate a list of data into a batch. This should take care of any desired
            padding.
        drop_last: Whether to drop the last batch of data if that batch is incomplete. Note that this is meaningless for
            batched datasets, as well as when `steps_per_epoch` is set - in which case the dataset will be re-sampled as
            necessary until the specified number of steps has been completed in full.
    """
    _current_threads = []
    FE_LOADER_KIND = 7

    # The typing for 'dataset' should be an 'and' rather than 'or' but that feature is still under development:
    # https://github.com/python/typing/issues/213

    def __init__(self,
                 dataset: Union[Dataset, Sized],
                 postprocess_fn: Optional[Callable[[Dict[str, Any]], Union[Dict[str, Any], FilteredData]]] = None,
                 batch_size: Optional[int] = 1,
                 steps_per_epoch: Optional[int] = None,
                 shuffle: bool = False,
                 num_workers: int = 0,
                 collate_fn: Callable = None,
                 drop_last: bool = False):
        reset_fn = dataset.fe_reset_ds if hasattr(dataset, 'fe_reset_ds') else None
        convert_fn = dataset.fe_batch_indices if hasattr(dataset, 'fe_batch_indices') else None
        sampler = InfiniteSampler(data_source=dataset, shuffle=shuffle, reset_fn=reset_fn, convert_fn=convert_fn)
        if batch_size is not None and batch_size < 1:
            raise ValueError(f"batch_size must be None or a positive integer, but got {batch_size}")
        # Figure out the real batch size. This is already done in OpDataset, but if user manually instantiates this
        # loader without using an OpDataset we still want to know the batch size
        if not hasattr(dataset, "fe_batch"):
            sample_item = dataset[0]
            dataset.fe_batch = len(sample_item) if isinstance(sample_item, list) else 0
        if dataset.fe_batch:
            # The batch size where torch is concerned is probably None, but we know that it is secretly batched
            self.fe_batch_size = dataset.fe_batch
        else:
            self.fe_batch_size = batch_size
        # Figure out how many samples should be returned during the course of 1 epoch
        if steps_per_epoch is not None:
            to_yield = steps_per_epoch * (batch_size or 1)
            # Note that drop_last is meaningless here since we will provide exactly the requested number of steps
        else:
            if isinstance(dataset, OpDataset) and isinstance(dataset.dataset, ExtendDataset):
                to_yield = dataset.dataset.spoof_length
            elif isinstance(dataset, ExtendDataset):
                to_yield = dataset.spoof_length
            else:
                to_yield = len(dataset)
            if drop_last:
                to_yield -= to_yield % (batch_size or 1)
        self.fe_samples_to_yield = to_yield
        self.fe_drop_last = drop_last
        self.fe_collate_fn = collate_fn or default_collate
        if self.fe_batch_size in (0, None) and batch_size is None and self.fe_collate_fn == default_collate:
            # The user did not provide a batch dataset nor a batch size, so default collate won't work. Have to try
            # convert instead.
            self.fe_collate_fn = default_convert
        self.fe_postprocess_fn = postprocess_fn

        # We could disable pre-collating when num_workers=0, but this would lead to inconsistent batch ordering between
        # single- and multi-processing.

        super().__init__(dataset=dataset,
                         batch_size=batch_size,
                         sampler=sampler,
                         num_workers=num_workers,
                         persistent_workers=False,
                         collate_fn=functools.partial(_pre_collate, try_fn=collate_fn, postprocess_fn=postprocess_fn),
                         worker_init_fn=lambda _: np.random.seed(random.randint(0, 2 ** 32 - 1)))
        if self.batch_size is not None:
            # We need a special fetcher type later in order to build batches correctly
            self._dataset_kind = self.FE_LOADER_KIND

    def shutdown(self) -> None:
        """Close the worker threads used by this iterator.

        The hope is that this will prevent "RuntimeError: DataLoader worker (pid(s) XXXX) exited unexpectedly" during
        the test suites.
        """
        if isinstance(self._iterator, _MultiProcessingDataLoaderIter):
            self._iterator._shutdown_workers()
        self._iterator = None
        FEDataLoader._current_threads.clear()

    def __iter__(self) -> _BaseDataLoaderIter:
        # Similar to the original iter method, but we remember iterators in order to manually close them when new ones
        # are created
        self.shutdown()
        self._iterator = self._get_fe_iterator()
        if isinstance(self._iterator, _MultiProcessingDataLoaderIter):
            FEDataLoader._current_threads.extend([w.pid for w in self._iterator._workers])
        return self._iterator

    def _get_fe_iterator(self):
        if self.num_workers == 0:
            if self.batch_size is None:
                # We use 'fake' batch size here to identify datasets which perform their own batching
                return _SPPostBatchIter(self)
            return _SPPreBatchIter(self)
        else:
            if self.batch_size is None:
                # We use 'fake' batch size here to identify datasets which perform their own batching
                return _MPPostBatchIter(self)
            return _MPPreBatchIter(self)


def _pre_collate(data: List[Union[FilteredData, Dict[str, Any]]],
                 try_fn: Callable[[List[Union[FilteredData, Dict[str, Any]]]], Dict[str, Any]],
                 postprocess_fn: Optional[Callable[[Dict[str, Any]], Union[Dict[str, Any], FilteredData]]]) -> \
        Tuple[Union[bool, FilteredData], Union[Dict[str, Any], List[Union[FilteredData, Dict[str, Any]]]]]:
    """A function which will try to pre-collate the data ahead of time in order to accelerate 'happy path'

    Args:
        data: An un-batched batch of data.
        try_fn: A collate function to attempt to use on the data.
        postprocess_fn: A function to run on batches of data after collation.

    Returns:
        Whether the data is batched, along with the data. If collate succeeded but the data was subsequently filtered,
        the first return value will be None.
    """
    try:
        collated = try_fn(data)
    except:
        # The presence of filtered data instances could break the possibly-user-specified collate function for any
        # reason, so cast a broad net on this except clause.
        return False, data
    if postprocess_fn is not None:
        collated = postprocess_fn(collated)
        if isinstance(collated, FilteredData):
            # There might be extra data sitting around that was supposed to be combined with this data, so return the
            # raw data and let the main thread try again.
            return collated, data
    return True, collated


class _BaseFELoaderIter(_BaseDataLoaderIter, ABC):
    """A base class for the FE Data iterators.

    Args:
        loader: The parent loader object that will own this iterator.
    """
    def __init__(self, loader: FEDataLoader):
        super().__init__(loader)
        self.fe_batch_size = loader.fe_batch_size
        self.fe_drop_last = loader.fe_drop_last
        self.fe_collate_fn = loader.fe_collate_fn
        self.fe_postprocess_fn = loader.fe_postprocess_fn
        self.fe_samples_to_yield = loader.fe_samples_to_yield
        self.fe_samples_yielded = 0
        self.fe_extra_data = []


def _next_pre_batch(self: _BaseFELoaderIter) -> Dict[str, Tensor]:
    """The __next__ function to use for a loader which is loading individual data instances (not a batchDataset)

    There is a failure mode with this logic where if the user artificially expands their dataset while using a filter,
    and if filtered data is extremely rare (like 1 data point per epoch), and that filtered point happens to get put in
    the same batch as point X multiple times, then the final batch of data may contain multiple copies of X.

    Args:
        self: The loader iterator onto which this method is grafted.

    Returns:
        A batch of data, collated into pytorch tensors.
    """
    if self._sampler_iter is None:
        self._reset()  # This copies the pytorch implementation, and seems to work, but is clearly missing an argument
    if self.fe_samples_yielded >= self.fe_samples_to_yield:
        self.fe_extra_data.clear()  # Throw out any old data
        raise StopIteration
    # Look for new data until you have enough to complete a batch
    while len(self.fe_extra_data) < self.fe_batch_size and self.fe_samples_yielded + len(
            self.fe_extra_data) < self.fe_samples_to_yield:
        sample_indices, (collated, candidate_batch) = self._next_data()
        if collated is True:  # not regular if check since collated might be FilteredData
            if self.fe_samples_yielded + self.fe_batch_size + len(self.fe_extra_data) < self.fe_samples_to_yield:
                # This batch can be used as-is
                self.fe_samples_yielded += self.fe_batch_size
                return candidate_batch
            # We are at the last batch of data for the epoch
            if not self.fe_extra_data and self.fe_samples_yielded + self.fe_batch_size == self.fe_samples_to_yield:
                # If we have no left-over data and the current batch is exactly the right size, then just use it
                self.fe_samples_yielded += self.fe_batch_size
                return candidate_batch
            # Otherwise we need to unpack this batch, combine it with any older data, and shrink it to the desired size
            # The easiest way to do this is to re-draw the batch from scratch since user might have run batched ops
            # which will have had invalid effects.
            candidate_batch = [self._dataset[idx] for idx in sample_indices]
        if isinstance(collated, FilteredData) and not self.fe_extra_data:
            # This batch was collated but filtered during a batch filter. It was returned here just in case there was
            # extra data lying around, but there isn't any. We can safely avoid processing it a second time.
            if not collated.replacement:
                self.fe_samples_yielded += min(len(candidate_batch), self.fe_samples_to_yield - self.fe_samples_yielded)
        else:
            for instance in candidate_batch:
                if self.fe_samples_yielded + len(self.fe_extra_data) == self.fe_samples_to_yield:
                    # Avoid processing extra data (for edge cases involving lots of filtered data near end of epoch)
                    break
                if isinstance(instance, FilteredData):
                    if not instance.replacement:
                        self.fe_samples_yielded += 1
                    continue
                self.fe_extra_data.append(instance)
    n_keep = min(self.fe_batch_size, self.fe_samples_to_yield - self.fe_samples_yielded)
    real_batch = self.fe_extra_data[:n_keep]
    self.fe_extra_data = self.fe_extra_data[n_keep:]
    self.fe_samples_yielded += len(real_batch)
    # Enforce drop_last if required
    if not real_batch or (self.fe_drop_last and len(real_batch) < self.fe_batch_size):
        self.fe_extra_data.clear()  # Throw out any extra data
        raise StopIteration
    # Collate the batch
    collated = self.fe_collate_fn(real_batch)
    # Apply any batch-level operations
    if self.fe_postprocess_fn is not None:
        collated = self.fe_postprocess_fn(collated)
        if isinstance(collated, FilteredData):
            if collated.replacement:
                self.fe_samples_yielded -= len(real_batch)
            return _next_pre_batch(self)
    return collated


def _next_post_batch(self: _BaseFELoaderIter) -> Dict[str, Tensor]:
    """The __next__ function to use for a loader which is loading batched data (ex. for a BatchDataset)

    Args:
        self: The loader iterator onto which this method is grafted.

    Returns:
        A batch of data, collated into pytorch tensors.
    """
    if self._sampler_iter is None:
        self._reset()  # This copies the pytorch implementation, and seems to work, but is clearly missing an argument
    while self.fe_samples_yielded < self.fe_samples_to_yield:
        collated, candidate_batch = self._next_data()
        if collated is True:  # Not regular if check since collated might be FilteredData
            self.fe_samples_yielded += 1
            return candidate_batch
        if isinstance(collated, FilteredData):
            # The batch was filtered during the forward_batch pass
            if not collated.replacement:
                self.fe_samples_yielded += 1
        else:
            # The filtered data appeared before batching, need to find it
            for instance in candidate_batch:
                if isinstance(instance, FilteredData):
                    if not instance.replacement:
                        self.fe_samples_yielded += 1
                    break
            else:
                # The else block is reached iff the for loop never breaks (probably should never get here)
                collated = self.fe_collate_fn(candidate_batch)
                if self.fe_postprocess_fn is not None:
                    collated = self.fe_postprocess_fn(collated)
                    if isinstance(collated, FilteredData):
                        if not collated.replacement:
                            self.fe_samples_yielded += 1
                        return _next_post_batch(self)
                self.fe_samples_yielded += 1
                return collated
    raise StopIteration


class _SPPreBatchIter(_BaseFELoaderIter, _SingleProcessDataLoaderIter):
    __next__ = _next_pre_batch


class _SPPostBatchIter(_BaseFELoaderIter, _SingleProcessDataLoaderIter):
    __next__ = _next_post_batch


class _MPPreBatchIter(_BaseFELoaderIter, _MultiProcessingDataLoaderIter):
    __next__ = _next_pre_batch


class _MPPostBatchIter(_BaseFELoaderIter, _MultiProcessingDataLoaderIter):
    __next__ = _next_post_batch


class InfiniteSampler(Sampler):
    """A class which never stops sampling.

    Args:
        data_source: The dataset to be sampled.
        shuffle: Whether to shuffle when sampling.
        reset_fn: A function to be invoked (using the provided `shuffle` arg) every time the dataset has been fully
            traversed.
        convert_fn: A function to be invoked (using the current index) every sample in order to convert an integer index
            into some arbitrary alternative index representation.
    """
    def __init__(self,
                 data_source: Sized,
                 shuffle: bool = True,
                 reset_fn: Optional[Callable[[bool], None]] = None,
                 convert_fn: Optional[Callable[[int], Any]] = None):
        super().__init__(data_source=data_source)
        self.ds_len = len(data_source)
        if self.ds_len < 1:
            raise ValueError("dataset length must be at least 1")
        self.indices = [i for i in range(self.ds_len)]
        self.shuffle = shuffle
        self.reset_fn = reset_fn
        self.convert_fn = convert_fn
        self.idx = 0

    def __len__(self):
        return self.ds_len

    def __iter__(self):
        self.idx = 0
        if self.reset_fn:
            self.reset_fn(self.shuffle)
        if self.shuffle:
            random.shuffle(self.indices)
        return self

    def __next__(self):
        if self.idx == self.ds_len:
            self.idx = 0
            if self.reset_fn:
                self.reset_fn(self.shuffle)
            if self.shuffle:
                random.shuffle(self.indices)
        elem = self.indices[self.idx]
        self.idx += 1
        if self.convert_fn:
            elem = self.convert_fn(elem)
        return elem


###
# Here we use a hack to patch a special fetcher into the pytorch ecosystem. This allows us to return the dataset indices
# alongside the collated data points in case the batch needs to be re-constructed later. It won't interfere with regular
# pytorch usage since our reserved 'kind' value is 7 whereas pytorch only uses 0 and 1.
###

class _IdxMapDatasetFetcher(_MapDatasetFetcher):
    def fetch(self, possibly_batched_index):
        if self.auto_collation:
            data = [self.dataset[idx] for idx in possibly_batched_index]
        else:
            data = self.dataset[possibly_batched_index]
        return possibly_batched_index, self.collate_fn(data)


if not hasattr(_DatasetKind, '_original_create_fetcher'):
    _DatasetKind._original_create_fetcher = _DatasetKind.create_fetcher

    def _create_fetcher(kind, dataset, auto_collation, collate_fn, drop_last):
        if kind == FEDataLoader.FE_LOADER_KIND:
            return _IdxMapDatasetFetcher(dataset, auto_collation, collate_fn, drop_last)
        else:
            return _DatasetKind._original_create_fetcher(kind, dataset, auto_collation, collate_fn, drop_last)

    _DatasetKind.create_fetcher = _create_fetcher



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
from collections import Counter
from fractions import Fraction
from typing import Iterable, List, Tuple, Union

import numpy as np
from nltk.translate.bleu_score import SmoothingFunction, brevity_penalty, modified_precision, sentence_bleu

from fastestimator.trace.meta.per_ds import per_ds
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_number


def get_formated_list(input_data: np.ndarray) -> List[str]:
    """
        Filter the padding(elements with 0 value) and typecast the elements of list to str.

        Returns:
            Formated list.
    """
    return [str(i) for i in input_data if i != 0]


def get_formated_reference(input_data: np.ndarray) -> List[List[str]]:
    """
        Encapsulate formated list in another list.

        Returns:
            List encapsulated formated list.
    """
    return [get_formated_list(input_data)]


@per_ds
@traceable()
class BleuScore(Trace):
    """Calculate the Bleu score for a nlp task and report it back to the logger.

       Calculate BLEU score (Bilingual Evaluation Understudy) from Papineni, Kishore, Salim Roukos, Todd Ward, and
       Wei-Jing Zhu. 2002."BLEU: a method for automatic evaluation of machine translation."In Proceedings of ACL.
       https://www.aclweb.org/anthology/P02-1040.pdf

       The BLEU metric scores a translation on a scale of 0 to 1, in an attempt to measure the adequacy and fluency of
       the Machine Translation output. The closer to 1 the test sentences score, the more overlap there is with their
       human reference translations and thus, the better the system is deemed to be. The MT output would score 1 only
       if it is identical to the reference human translation. But even two competent human translations of the exact
       same material may only score in the 0.6 or 0.7 range as they are likely to use different vocabulary and phrasing.
       We should be wary of very high BLEU scores (in excess of 0.7) as it is probably measuring improperly or overfitting.

       The default BLEU calculates a score for up to 4-grams using uniform weights (this is called BLEU-4). To evaluate
       your translations with lower order ngrams, use customized "n_gram". E.g. when accounting for up to 2-grams
       with uniform weights (this is called BLEU-2) use n_gram=2.

       If there is no ngrams overlap for any order of n-grams, BLEU returns the value 0. This is because the precision
       for the order of n-grams withoutoverlap is 0, and the geometric mean in the final BLEU score computation multiplies
       the 0 with the precision of other n-grams. This results in 0. Shorter translations may have inflated precision values due to having
       smaller denominators; therefore, we give them proportionally smaller smoothed counts. Instead of scaling to 1/(2^k),
       Chen and Cherry suggests dividing by 1/ln(len(T)), where T is the length of the translation.


    Args:
        true_key: Name of the key that corresponds to ground truth in the batch dictionary.
        pred_key: Name of the key that corresponds to predicted score in the batch dictionary.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        output_name: Name of the key to store back to the state.
        n_gram: Number of grams used to calculate bleu score.
        per_ds: Whether to automatically compute this metric individually for every ds_id it runs on, in addition to
            computing an aggregate across all ds_ids on which it runs. This is automatically False if `output_name`
            contains a "|" character.

    """
    def __init__(self,
                 true_key: str,
                 pred_key: str,
                 mode: Union[None, str, Iterable[str]] = ("eval", "test"),
                 ds_id: Union[None, str, Iterable[str]] = None,
                 output_name: str = 'bleu_score',
                 n_gram: int = 4,
                 per_ds: bool = True) -> None:
        super().__init__(inputs=(true_key, pred_key), outputs=output_name, mode=mode, ds_id=ds_id)
        self.n_gram = n_gram
        self.weights = self.get_output_weights()
        self.per_ds = per_ds
        self.smoothing_function = SmoothingFunction().method4
        self.no_of_correct_predicted = Counter()
        self.no_of_total_predicted = Counter()
        self.total_hypotheses_length = 0
        self.total_references_length = 0

    @property
    def true_key(self) -> str:
        return self.inputs[0]

    @property
    def pred_key(self) -> str:
        return self.inputs[1]

    def get_output_weights(self) -> Tuple[float, ...]:
        """
            Generate weights tuple based on n_gram.

            Returns:
                Tuple of n_gram weights

            Raises:
                ValueError: When n_gram provided is less than or equal to 0..
        """
        if self.n_gram > 0:
            return (1 / self.n_gram, ) * self.n_gram
        else:
            raise ValueError("N Gram should be a positive integer.")

    def on_epoch_begin(self, data: Data) -> None:
        self.no_of_correct_predicted = Counter()
        self.no_of_total_predicted = Counter()
        self.total_hypotheses_length = 0
        self.total_references_length = 0

    def get_brevity_penalty(self) -> float:
        """
            Calculate the brevity penalty of the corpus.

            Returns:
                Brevity penalty for corpus.
        """
        return brevity_penalty(self.total_references_length, self.total_hypotheses_length)

    def get_smoothened_modified_precision(self) -> List[float]:
        """
            Calculate the smoothened modified precision.

            Returns:
                List of smoothened modified precision of n_grams.
        """
        # Collects the various precision values for the different ngram orders.
        p_n = [
            Fraction(self.no_of_correct_predicted[i], self.no_of_total_predicted[i], _normalize=False)
            for i in range(1, self.n_gram + 1)
        ]
        # Smoothen the modified precision.
        return self.smoothing_function(p_n, [], [], hyp_len=self.total_hypotheses_length)

    def get_corpus_bleu_score(self) -> float:
        """
            Calculate the bleu score using corpus level brevity penalty and geometric average precision.

            Returns:
                Corpus level bleu score.
        """
        # Calculate corpus-level brevity penalty.
        bp = self.get_brevity_penalty()

        # Returns 0 if there's no matching 1-gram
        if self.no_of_correct_predicted[1] == 0:
            return 0

        n_gram_precision = self.get_smoothened_modified_precision()

        geometric_average_precision = math.exp(
            math.fsum((w_i * math.log(p_i) for w_i, p_i in zip(self.weights, n_gram_precision) if p_i > 0)))
        bleu_score = bp * geometric_average_precision

        return bleu_score

    def batch_precision_parameters(self, references: List[np.ndarray], hypotheses: List[np.ndarray]) -> List[float]:
        """
            Calculate modified precision per n_gram for input references and hypotheses combinations.

            Args:
                references: Ground truth sentences.
                hypotheses: Predicted sentences.

            Returns:
                List of sentence level bleu scores
        """

        assert len(references) == len(hypotheses), (
            "The number of hypotheses and their reference(s) should be the same ")

        sentence_level_scores = []
        # Iterate through each hypothesis and their corresponding references.
        for reference, hypothesis in zip(references, hypotheses):

            # For each order of ngram, calculate the correct predicted words and
            # total predicted words for the corpus-level modified precision.
            reference = get_formated_reference(reference)
            hypothesis = get_formated_list(hypothesis)
            for i in range(1, self.n_gram + 1):
                p_i = modified_precision(reference, hypothesis, i)
                self.no_of_correct_predicted[i] += p_i.numerator
                self.no_of_total_predicted[i] += p_i.denominator

            sentence_level_scores.append(sentence_bleu(reference, hypothesis, self.weights, self.smoothing_function))

            # Calculate the hypothesis length and the closest reference length.
            # Adds them to the corpus-level hypothesis and reference counts.
            hyp_len = len(hypothesis)
            self.total_hypotheses_length += hyp_len
            ref_lens = (len(ref) for ref in reference)
            self.total_references_length += min(ref_lens, key=lambda ref_len: (abs(ref_len - hyp_len), ref_len))

        return sentence_level_scores

    def on_batch_end(self, data: Data) -> None:
        y_pred, y_true = to_number(data['pred']), to_number(data['target_real'])
        if y_true.shape[-1] > 1 and y_true.ndim > 2:
            y_true = np.argmax(y_true, axis=-1)
        if y_pred.shape[-1] > 1 and y_pred.ndim > 2:
            y_pred = np.argmax(y_pred, axis=-1)
        sentence_level_scores = self.batch_precision_parameters(y_true, y_pred)
        data.write_per_instance_log(self.outputs[0], sentence_level_scores)

    def on_epoch_end(self, data: Data) -> None:
        data.write_with_log(self.outputs[0], round(self.get_corpus_bleu_score(), 5))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from torch.utils.data import Dataset

from fastestimator.util.traceability_util import traceable


@traceable()
class ExtendDataset(Dataset):
    """ExtendDataset either extends or contracts the length of provided Dataset.

    Args:
        dataset: The Original dataset(s) which need expansion or contraction.
        spoof_length: Length to which original dataset must be expanded or contracted to. (New desired length)
    """
    def __init__(self, dataset: Dataset, spoof_length: int) -> None:
        self.dataset = dataset
        self.spoof_length = spoof_length
        self._check_input()
        if hasattr(dataset, "fe_reset_ds"):
            self.fe_reset_ds = dataset.fe_reset_ds
        if hasattr(dataset, "fe_batch_indices"):
            self.fe_batch_indices = dataset.fe_batch_indices
        if hasattr(dataset, "fe_batch"):
            self.fe_batch = dataset.fe_batch

    def __len__(self):
        return len(self.dataset)

    def _check_input(self) -> None:
        """Verify that the given input values are valid.
        Raises:
            AssertionError: If any of the parameters are found to by unacceptable for a variety of reasons.
        """
        assert isinstance(self.spoof_length, int), "Only accept positive integer type as spoof_length"
        assert self.spoof_length > 0, "Invalid spoof_length. Expand Length cannot be less than or equal to 0"
        assert not isinstance(self.dataset, ExtendDataset), "Input Dataset cannot be an ExtendDataset object"

    def __getitem__(self, index):
        return self.dataset[index]



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf

from fastestimator.backend import to_tensor
from fastestimator.backend.tensor_normalize import normalize


class TestNormalize(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.numpy_array = np.arange(0.0, 12.0, 1.0, dtype=np.float32).reshape((1, 2, 2, 3))
        self.numpy_array_int = np.arange(0.0, 12.0, 1.0, dtype=int).reshape((1, 2, 2, 3))
        self.expected_result = np.array([[[[-1.593255, -1.3035723, -1.0138896], [-0.7242068, -0.4345241, -0.14484136]],
                                          [[0.14484136, 0.4345241, 0.7242068], [1.0138896, 1.3035723, 1.593255]]]],
                                        dtype=np.float32)

    def test_normalize_np_value(self):
        np.testing.assert_array_almost_equal(normalize(self.numpy_array, 0.5, 0.31382295, 11.0), self.expected_result)

    def test_normalize_np_value_int(self):
        np.testing.assert_array_almost_equal(normalize(self.numpy_array_int, 0.5, 0.31382295, 11), self.expected_result)

    def test_normalize_tf_value(self):
        np.testing.assert_array_almost_equal(
            normalize(tf.convert_to_tensor(self.numpy_array), 0.5, 0.31382295, 11.0).numpy(), self.expected_result)

    def test_normalize_tf_value_int(self):
        np.testing.assert_array_almost_equal(
            normalize(tf.convert_to_tensor(self.numpy_array_int), 0.5, 0.31382295, 11.0).numpy(), self.expected_result)

    def test_normalize_torch_value(self):
        np.testing.assert_array_almost_equal(
            normalize(to_tensor(self.numpy_array, 'torch'), 0.5, 0.31382295, 11.0).numpy(), self.expected_result)

    def test_normalize_torch_value_int(self):
        np.testing.assert_array_almost_equal(
            normalize(to_tensor(self.numpy_array_int, 'torch'), 0.5, 0.31382295, 11.0).numpy(), self.expected_result)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar, Union

import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


def l2_regularization(model: Union[tf.keras.Model, torch.nn.Module], beta: float = 0.01) -> Tensor:
    """Calculate L2 Norm of model weights.

    l2_reg = sum(parameter**2)/2

    This method can be used with TensorFlow and Pytorch tensors

    Args:
        model: A tensorflow or pytorch model
        beta: The multiplicative factor, to weight the l2 regularization loss with the input loss

    Returns:
        The L2 norm of model parameters

    Raises:
        ValueError: If `model` belongs to an unacceptable framework.
    """
    if isinstance(model, torch.nn.Module):
        l2_loss = torch.sum(torch.stack([torch.sum(p**2) / 2 for p in model.parameters() if p.requires_grad]))
    elif isinstance(model, tf.keras.Model):
        l2_loss = tf.reduce_sum([tf.nn.l2_loss(p) for p in model.trainable_variables])
    else:
        raise ValueError("Unrecognized model framework: Please make sure to pass either torch or tensorflow models")
    return beta * l2_loss



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from fastestimator.dataset.numpy_dataset import NumpyDataset
from sklearn import datasets


def load_data(image_key: str = "x", label_key: str = "y") -> NumpyDataset:
    """Load and return the Sklearn digits dataset.

    Args:
        image_key: The key for image.
        label_key: The key for label.

    Returns:
        (train_data, eval_data)
    """
    ds = datasets.load_digits()
    images = ds.images
    targets = ds.target
    return NumpyDataset({image_key: images, label_key: targets})



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Callable, Dict, Iterable, List, MutableMapping, Optional, TypeVar, Union

import numpy as np
import tensorflow as tf
import torch
from torch.utils.data.dataloader import default_collate

from fastestimator.backend.to_tensor import to_tensor
from fastestimator.op.op import Op, get_inputs_by_op, write_outputs_by_op
from fastestimator.util.data import FilteredData
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import pad_batch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


@traceable()
class NumpyOp(Op):
    """An Operator class which takes and returns numpy data.

    These Operators are used in fe.Pipeline to perform data pre-processing / augmentation. They may also be used in
    fe.Network to perform postprocessing on data.

    Args:
        inputs: Key(s) from which to retrieve data from the data dictionary.
        outputs: Key(s) under which to write the outputs of this Op back to the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 inputs: Union[None, str, Iterable[str]] = None,
                 outputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        # in_place_edits tracks whether the .forward() method of this op will perform in-place edits of numpy arrays.
        # This is inferred automatically by the system and is used for memory management optimization. If you are
        # developing a NumpyOp which does in-place edits, the best practice is to set this to True in your init method.
        self.in_place_edits = False

    def forward(self, data: Union[np.ndarray, List[np.ndarray]],
                state: Dict[str, Any]) -> Union[None, FilteredData, np.ndarray, List[np.ndarray]]:
        """A method which will be invoked in order to transform data.

        This method will be invoked on individual elements of data before any batching / axis expansion is performed.

        Args:
            data: The arrays from the data dictionary corresponding to whatever keys this Op declares as its `inputs`.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The `data` after applying whatever transform this Op is responsible for. It will be written into the data
            dictionary based on whatever keys this Op declares as its `outputs`.
        """
        return data

    def forward_batch(self,
                      data: Union[np.ndarray, List[np.ndarray]],
                      state: Dict[str, Any]) -> Union[None, FilteredData, np.ndarray, List[np.ndarray]]:
        """A method which will be invoked in order to transform a batch of data.

        This method will be invoked on batches of data during network postprocessing. It should expect to receive
        batched data and should itself return batched data.

        Args:
            data: The arrays from the data dictionary corresponding to whatever keys this Op declares as its `inputs`.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The `data` after applying whatever transform this Op is responsible for. It will be written into the data
            dictionary based on whatever keys this Op declares as its `outputs`.
        """
        if isinstance(data, list):
            data = [elem for elem in map(list, zip(*data))]
        else:
            data = [elem for elem in data]
        results = [self.forward(elem, state) for elem in data]
        if self.out_list:
            results = [np.array(col) for col in [[row[i] for row in results] for i in range(len(results[0]))]]
        else:
            results = np.array(results)
        return results


@traceable()
class Batch(NumpyOp):
    """Convert data instances into a batch of data.

    Only one instance of a Batch Op can be present for a given epoch/mode/ds_id combination. Any Ops after this one will
    operate on batches of data rather than individual instances (using their batch_forward methods).

    Args:
        batch_size: The batch size to use. If set, this will override any value specified by the Pipeline, allowing
            control of the batch size on a per-mode and per-ds_id level. Note that this value will be ignored when using
            a BatchDataset (or any dataset which decides on its own batch configuration).
        drop_last: Whether to drop the last batch if the last batch is incomplete. Note that setting this to True when
            using a BatchDataset (or any dataset which decides on its own batch configuration) won't do anything.
        pad_value: The padding value if batch padding is needed. None indicates that no padding is needed. Mutually
            exclusive with `collate_fn`.
        collate_fn: A function to merge a list of data elements into a batch of data. Mutually exclusive with
            `pad_value`.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 batch_size: Optional[int] = None,
                 drop_last: bool = False,
                 pad_value: Optional[Union[int, float]] = None,
                 collate_fn: Optional[Callable[[List[Dict[str, Any]]], Dict[str, Any]]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(mode=mode, ds_id=ds_id)
        if batch_size is not None:
            if not isinstance(batch_size, int):
                raise ValueError(f"batch_size must be an integer, but got {type(batch_size)}")
            if batch_size < 0:
                raise ValueError("batch_size must be non-negative")
        self.batch_size = batch_size
        self.drop_last = drop_last
        if pad_value is not None and collate_fn is not None:
            raise ValueError("Provide either a pad_value or collate_fn, but not both")
        self._pad_value = pad_value
        self.collate_fn = collate_fn
        if pad_value is not None:
            self.collate_fn = self._pad_batch_collate
        if self.collate_fn is None:
            # Note that this might get ignored in favor of default_convert inside the FEDataLoader if it looks like the
            # user really doesn't want stuff to be batched.
            self.collate_fn = default_collate

    def _pad_batch_collate(self, batch: List[MutableMapping[str, Any]]) -> Dict[str, Any]:
        """A collate function which pads a batch of data.

        Args:
            batch: The data to be batched and collated.

        Returns:
            A padded and collated batch of data.
        """
        pad_batch(batch, self._pad_value)
        return default_collate(batch)


@traceable()
class Delete(NumpyOp):
    """Delete key(s) and their associated values from the data dictionary.

    The system has special logic to detect instances of this Op and delete its `inputs` from the data dictionary.

    Args:
        keys: Existing key(s) to be deleted from the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 keys: Union[str, List[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(inputs=keys, mode=mode, ds_id=ds_id)

    def forward(self, data: Union[np.ndarray, List[np.ndarray]], state: Dict[str, Any]) -> None:
        pass

    def forward_batch(self, data: Union[Tensor, List[Tensor]], state: Dict[str, Any]) -> None:
        pass


@traceable()
class LambdaOp(NumpyOp):
    """An Operator that performs any specified function as forward function.

    Args:
        fn: The function to be executed.
        inputs: Key(s) from which to retrieve data from the data dictionary.
        outputs: Key(s) under which to write the outputs of this Op back to the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 fn: Callable,
                 inputs: Union[None, str, Iterable[str]] = None,
                 outputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.fn = fn
        self.in_list = True

    def set_rua_level(self, magnitude_coef: float) -> None:
        """A method which will be invoked by the RUA Op to adjust the augmentation intensity.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> Union[np.ndarray, List[np.ndarray]]:
        return self.fn(*data)

    def forward_batch(self, data: Union[Tensor, List[Tensor]], state: Dict[str,
                                                                           Any]) -> Union[np.ndarray, List[np.ndarray]]:
        return self.forward(data, state)


@traceable()
class RemoveIf(NumpyOp):
    """An Operator which will remove a datapoint from the pipeline if the given criterion is satisfied.

    Args:
        fn: A function taking any desired `inputs` and returning a boolean. If the return value is true, the current
            datapoint (or batch if using a batch dataset) will be removed and another will take its place.
        replacement: Whether to replace the filtered element with another (thus maintaining the number of steps in an
            epoch but potentially increasing data repetition) or else shortening the epoch by the number of filtered
            data points (fewer steps per epoch than expected, but no extra data repetition). Either way, the number of
            data points within an individual batch will remain the same. Even if `replacement` is true, data will not be
            repeated until all of the given epoch's data has been traversed (except for at most 1 batch of data which
            might not appear until after the re-shuffle has occurred).
        inputs: Key(s) from which to retrieve data from the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 fn: Callable[..., bool],
                 replacement: bool = True,
                 inputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(inputs=inputs, mode=mode, ds_id=ds_id)
        self.filter_fn = fn
        self.in_list = True
        self.replacement = replacement

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> Optional[FilteredData]:
        if self.filter_fn(*data):
            return FilteredData(replacement=self.replacement)
        return None

    def forward_batch(self, data: Union[Tensor, List[Tensor]], state: Dict[str,
                                                                           Any]) -> Optional[FilteredData]:
        return self.forward(data, state)


def forward_numpyop(ops: List[NumpyOp],
                    data: MutableMapping[str, Any],
                    state: Dict[str, Any],
                    batched: Optional[str] = None) -> Optional[FilteredData]:
    """Call the forward function for list of NumpyOps, and modify the data dictionary in place.

    Args:
        ops: A list of NumpyOps to execute.
        data: The data dictionary.
        state: Information about the current execution context, ex. {"mode": "train"}. Must contain at least the mode.
        batched: Whether the `data` is batched or not. If it is batched, provide the string ('tf', 'torch', or 'np')
            indicating which type of tensors the batch contains.
    """
    if batched:
        # Cast data to Numpy before performing batch forward
        for key, val in data.items():
            data[key] = to_tensor(val, target_type='np')
    for op in ops:
        op_data = get_inputs_by_op(op, data, copy_on_write=op.in_place_edits)
        try:
            op_data = op.forward_batch(op_data, state) if batched else op.forward(op_data, state)
        except ValueError as err:
            if err.args[0] == 'assignment destination is read-only':
                # If the numpy error text changes we'll need to make adjustments in the future
                op.in_place_edits = True
                op_data = get_inputs_by_op(op, data, copy_on_write=op.in_place_edits)
                op_data = op.forward_batch(op_data, state) if batched else op.forward(op_data, state)
            else:
                raise err
        if isinstance(op_data, FilteredData):
            return op_data
        if isinstance(op, Delete):
            for key in op.inputs:
                del data[key]
        if op.outputs:
            write_outputs_by_op(op, data, op_data)
    if batched:
        # Cast data back to original tensor type after performing batch forward
        for key, val in data.items():
            data[key] = to_tensor(val, target_type=batched, shared_memory=True)
    return None



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import zipfile
from pathlib import Path
from typing import Optional, Tuple

import wget

from fastestimator.dataset.labeled_dir_dataset import LabeledDirDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def load_data(root_dir: Optional[str] = None) -> Tuple[LabeledDirDataset, LabeledDirDataset]:
    """Load and return the Tiny ImageNet dataset.
    Sourced from http://cs231n.stanford.edu/tiny-imagenet-200.zip. This method will
        download the data to local storage if the data has not been previously downloaded.
    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.
    Returns:
        (train_data, eval_data)
    """
    home = str(Path.home())

    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'tiny_imagenet')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'tiny_imagenet')
    os.makedirs(root_dir, exist_ok=True)

    data_compressed_path = os.path.join(root_dir, 'tiny-imagenet-200.zip')
    train_file_path = os.path.join(root_dir, 'tiny-imagenet-200', 'train')
    val_file_path = os.path.join(root_dir, 'tiny-imagenet-200', 'val')

    if (os.path.exists(train_file_path) == False) or (os.path.exists(val_file_path) == False):
        # download
        if not os.path.exists(data_compressed_path):
            print("Downloading data to {}".format(root_dir))
            wget.download('http://cs231n.stanford.edu/tiny-imagenet-200.zip', root_dir, bar=bar_custom)

        # extract
        print("\nExtracting files ...")
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            zip_file.extractall(root_dir)
        #os.rename(os.path.join(root_dir, 'horse2zebra'), data_folder_path)

        current_dir = os.path.join(root_dir, 'tiny-imagenet-200')

        # Update Train Directory
        for root, _, files in os.walk(os.path.join(current_dir, 'train')):
            for filename in files:
                #print(os.path.join(root, filename))
                if filename.endswith('.txt'):
                    os.remove(os.path.join(root, filename))
                else:
                    p = Path(os.path.join(root, filename)).absolute()
                    parent_dir = p.parents[1]
                    p.rename(parent_dir / p.name)
            if len(os.listdir(root)) == 0:
                os.rmdir(root)

        # Update Val Directory
        for line in open(os.path.join(current_dir, 'val', 'val_annotations.txt')).readlines():
            file_data = [n for n in line.split()]
            folder_path = os.path.join(current_dir, 'val', file_data[1])
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)

            file_path = os.path.join(current_dir, 'val', 'images', file_data[0])
            os.rename(file_path, os.path.join(folder_path, file_data[0]))
        os.rmdir(os.path.join(current_dir, 'val', 'images'))
        os.remove(os.path.join(current_dir, 'val', 'val_annotations.txt'))

    root_path = os.path.join(root_dir, 'tiny-imagenet-200')
    train_outputs = LabeledDirDataset(os.path.join(root_path, "train"), data_key='image', label_key='label')

    eval_outputs = LabeledDirDataset(os.path.join(root_path, "val"), data_key='image', label_key='label')

    return (train_outputs, eval_outputs)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.l2_regularization import l2_regularization
from fastestimator.op.tensorop.tensorop import TensorOp

from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


@traceable()
class L2Regularizaton(TensorOp):
    """Calculate L2 Regularization Loss.

    Args:
        inputs: String key representing input loss.
        outputs: String key under which to store the computed loss value.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        model: A tensorflow or pytorch model
        beta: The multiplicative factor, to weight the l2 regularization loss with the input loss
    """
    def __init__(self,
                 inputs: str,
                 outputs: str,
                 model: Union[tf.keras.Model, torch.nn.Module],
                 mode: Union[None, str, Iterable[str]] = None,
                 beta: float = 0.01):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)

        self.model = model
        self.beta = beta

    def forward(self, data: Union[Tensor, List[Tensor]], state: Dict[str, Any]) -> Tensor:
        loss = data
        total_loss = l2_regularization(self.model, self.beta) + loss
        return total_loss



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, ChainMap, Dict, List, MutableMapping, Optional


class Data(ChainMap[str, Any]):
    """A class which contains prediction and batch data.

    This class is intentionally not @traceable.

    Data objects can be interacted with as if they are regular dictionaries. They are however, actually a combination of
    two dictionaries, a dictionary for trace communication and a dictionary of prediction+batch data. In general, data
    written into the trace dictionary will be logged by the system, whereas data in the pred+batch dictionary will not.
    We therefore provide helper methods to write entries into `Data` which are intended or not intended for logging.

    ```python
    d = fe.util.Data({"a":0, "b":1, "c":2})
    a = d["a"]  # 0
    d.write_with_log("d", 3)
    d.write_without_log("e", 5)
    d.write_with_log("a", 4)
    a = d["a"]  # 4
    r = d.read_logs(extra_keys={"c"})  # {"c":2, "d":3, "a":4}
    ```

    Args:
        batch_data: The batch data dictionary. In practice this is itself often a ChainMap containing separate
            prediction and batch dictionaries.
    """
    maps: List[MutableMapping[str, Any]]

    def __init__(self, batch_data: Optional[MutableMapping[str, Any]] = None) -> None:
        super().__init__({}, batch_data or {}, {})
        self.per_instance_enabled = True  # Can be toggled if you need to block traces from recording detailed info

    def write_with_log(self, key: str, value: Any) -> None:
        """Write a given `value` into the `Data` dictionary with the intent that it be logged.

        Args:
            key: The key to associate with the new entry.
            value: The new entry to be written.
        """
        self.__setitem__(key, value)

    def write_without_log(self, key: str, value: Any) -> None:
        """Write a given `value` into the `Data` dictionary with the intent that it not be logged.

        Args:
            key: The key to associate with the new entry.
            value: The new entry to be written.
        """
        self.maps[1][key] = value

    def write_per_instance_log(self, key: str, value: Any) -> None:
        """Write a given per-instance `value` into the `Data` dictionary for use with detailed loggers (ex. CSVLogger).

        Args:
            key: The key to associate with the new entry.
            value: The new per-instance entry to be written.
        """
        if self.per_instance_enabled:
            self.maps[2][key] = value

    def read_logs(self) -> MutableMapping[str, Any]:
        """Read all values from the `Data` dictionary which were intended to be logged.

        Returns:
            A dictionary of all of the keys and values to be logged.
        """
        return self.maps[0]

    def read_per_instance_logs(self) -> MutableMapping[str, Any]:
        """Read all per-instance values from the `Data` dictionary for detailed logging.

        Returns:
             A dictionary of the keys and values to be logged.
        """
        return self.maps[2]


class DSData(Data):
    # noinspection PyMissingConstructor
    def __init__(self, ds_id: str, data: Data):
        self.maps = data.maps
        self.ds_id = ds_id
        self.per_instance_enabled = True

    def write_with_log(self, key: str, value: Any) -> None:
        super().write_with_log(key=f'{key}|{self.ds_id}', value=value)

    def write_without_log(self, key: str, value: Any) -> None:
        super().write_without_log(key=f'{key}|{self.ds_id}', value=value)

    def write_per_instance_log(self, key: str, value: Any) -> None:
        super().write_per_instance_log(key=f'{key}|{self.ds_id}', value=value)


class FilteredData:
    """A placeholder to indicate that this data instance should not be used.

    This class is intentionally not @traceable.

    Args:
        replacement: Whether to replace the filtered element with another (thus maintaining the number of steps in an
            epoch but potentially increasing data repetition) or else shortening the epoch by the number of filtered
            data points (fewer steps per epoch than expected, but no extra data repetition). Either way, the number of
            data points within an individual batch will remain the same. Even if `replacement` is true, data will not be
            repeated until all of the given epoch's data has been traversed (except for at most 1 batch of data which
            might not appear until after the re-shuffle has occurred).
    """
    def __init__(self, replacement: bool = True):
        self.replacement = replacement

    def __repr__(self):
        return "FilteredData"



# Copyright 2022 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend import to_tensor
from fastestimator.op.tensorop.normalize import Normalize


class TestNormalize(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.numpy_array = np.arange(0.0, 27.0, 1.0, dtype=np.float32).reshape((1, 3, 3, 3))
        self.numpy_array_int = np.arange(0, 27, 1, dtype=int).reshape((1, 3, 3, 3))
        self.expected_result = np.array(
            [[[[-1.6688062, -1.5404365, -1.4120668], [-1.283697, -1.1553273, -1.0269576], [
                -0.89858794, -0.77021825, -0.6418485
            ]], [[-0.5134788, -0.38510913, -0.2567394], [-0.1283697, 0., 0.1283697], [0.2567394, 0.38510913, 0.5134788]
                 ], [[0.6418485, 0.77021825, 0.89858794], [1.0269576, 1.1553273, 1.283697],
                     [1.4120668, 1.5404365, 1.6688062]]]],
            dtype=np.float32)
        self.expected_result_multi = np.array(
            [[[[-1.5331011, -1.543425, -1.5537487], [-1.1459544, -1.1562783, -1.166602],
               [-0.7588076, -0.7691315, -0.7794553]],
              [[-0.37166086, -0.38198477, -0.39230856], [0.01548585, 0.00516195, -0.00516183], [
                  0.4026326, 0.39230868, 0.3819849
              ]], [[0.7897793, 0.7794554, 0.7691316], [1.176926, 1.1666021, 1.1562784],
                   [1.5640727, 1.5537488, 1.5434251]]]],
            dtype=np.float32)

    def test_normalize_tf_int(self):
        op = Normalize(inputs="image", outputs="image", mean=0.482, std=0.289, max_pixel_value=27)
        data = op.forward(data=tf.convert_to_tensor(self.numpy_array), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result, 2)

    def test_normalize_tf_multi_int(self):
        op = Normalize(inputs="image",
                       outputs="image",
                       mean=(0.44, 0.48, 0.52),
                       std=(0.287, 0.287, 0.287),
                       max_pixel_value=27)
        data = op.forward(data=tf.convert_to_tensor(self.numpy_array), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result_multi, 2)

    def test_normalize_torch(self):
        op = Normalize(inputs="image", outputs="image", mean=0.482, std=0.289, max_pixel_value=27.0)
        data = op.forward(data=to_tensor(self.numpy_array, "torch"), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result, 2)

    def test_normalize_torch_multi(self):
        op = Normalize(inputs="image",
                       outputs="image",
                       mean=(0.44, 0.48, 0.52),
                       std=(0.287, 0.287, 0.287),
                       max_pixel_value=27)
        data = op.forward(data=to_tensor(self.numpy_array, "torch"), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result_multi, 2)

    def test_normalize_torch_float(self):
        op = Normalize(inputs="image", outputs="image", mean=0.482, std=0.289, max_pixel_value=27.0)
        data = op.forward(data=to_tensor(self.numpy_array, "torch"), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result, 2)

    def test_normalize_torch_multi_float(self):
        op = Normalize(inputs="image",
                       outputs="image",
                       mean=(0.44, 0.48, 0.52),
                       std=(0.287, 0.287, 0.287),
                       max_pixel_value=27)
        data = op.forward(data=to_tensor(self.numpy_array, "torch"), state={})
        np.testing.assert_array_almost_equal(data.numpy(), self.expected_result_multi, 2)

    def test_normalize_numpy_float(self):
        op = Normalize(inputs="image", outputs="image", mean=0.482, std=0.289, max_pixel_value=27.0)
        data = op.forward(data=self.numpy_array, state={})
        np.testing.assert_array_almost_equal(data, self.expected_result, 2)

    def test_normalize_numpy_multi_float(self):
        op = Normalize(inputs="image",
                       outputs="image",
                       mean=(0.44, 0.48, 0.52),
                       std=(0.287, 0.287, 0.287),
                       max_pixel_value=27)
        data = op.forward(data=self.numpy_array, state={})
        np.testing.assert_array_almost_equal(data, self.expected_result_multi, 2)
